""" define names, clean names and values
autoname adds geometric hash

```
import importlib
function_string = c.get_settings()['module']
mod_name, func_name = function_string.rsplit('.',1)
mod = importlib.import_module(mod_name)
func = getattr(mod, func_name)
result = func()
```
"""
import functools
from inspect import signature
import hashlib
import numpy as np
from phidl import Device

MAX_NAME_LENGTH = 32


def join_first_letters(name):
    """ join the first letter of a name separated with underscores (taper_length -> TL) """
    return "".join([x[0] for x in name.split("_") if x])


def get_component_name(component_type, max_name_length=MAX_NAME_LENGTH, **kwargs):
    name = component_type

    if kwargs:
        name += "_" + dict2name(**kwargs)

    # If the name is too long, fall back on hashing the longuest arguments
    if len(name) > max_name_length:
        shorter_name = f"{component_type}_{hashlib.md5(name.encode()).hexdigest()[:8]}"
        # print(f"{name} -> {shorter_name} ({len(name)} -> {max_name_length})")
        name = shorter_name

    return name


def autoname(component_function):
    """ decorator for auto-naming component functions
    if no Keyword argument `name`  is passed it creates a name by concenating all Keyword arguments

    .. plot::
      :include-source:

      import pp

      @pp.autoname
      def rectangle(size=(4,2), layer=0):
          c = pp.Component()
          w, h = size
          points = [[w, h], [w, 0], [0, 0], [0, h]]
          c.add_polygon(points, layer=layer)
          return c

      c = rectangle(layer=1)
      c << pp.c.text(text=c.name, size=1)
      pp.plotgds(c)

    """

    @functools.wraps(component_function)
    def wrapper(*args, **kwargs):
        if args:
            raise ValueError("autoname supports only Keyword args")
        max_name_length = kwargs.pop("max_name_length", MAX_NAME_LENGTH)
        name = kwargs.pop(
            "name",
            get_component_name(
                component_function.__name__, max_name_length=max_name_length, **kwargs
            ),
        )
        kwargs.pop("ignore_from_name", [])

        component = component_function(**kwargs)
        component.name = name
        component.module = component_function.__module__
        component.function_name = component_function.__name__
        sig = signature(component_function)
        component.settings.update(
            **{p.name: p.default for p in sig.parameters.values()}
        )
        component.settings.update(**kwargs)
        # if hasattr(component, 'hash_geometry'):
        #     component.settings.update(hash=component.hash_geometry())

        return component

    return wrapper


def dict2name(prefix=None, **kwargs):
    """ returns name from a dict """
    ignore_from_name = kwargs.pop("ignore_from_name", [])

    if prefix:
        label = [prefix]
    else:
        label = []
    for key in sorted(kwargs):
        if key not in ignore_from_name:
            value = kwargs[key]
            key = join_first_letters(key)
            value = clean_value(value)
            label += [f"{key.upper()}{value}"]
    label = "_".join(label)
    return clean_name(label)


def clean_name(name):
    """ Ensures that gds cells are composed of [a-zA-Z0-9]

    FIXME: only a few characters are currently replaced.
        This function has been updated only on case-by-case basis
    """
    replace_map = {
        " ": "_",
        "!": "_",
        "#": "_",
        "%": "_",
        "(": "",
        ")": "",
        "*": "_",
        ",": "_",
        "-": "m",
        ".": "p",
        "/": "_",
        ":": "_",
        "=": "",
        "@": "_",
        "[": "",
        "]": "",
    }
    for k, v in list(replace_map.items()):
        name = name.replace(k, v)
    return name


def clean_value(value):
    """ returns more readable value (integer)
    if number is < 1:
        returns number units in nm (integer)
    """

    try:
        if isinstance(value, int):  # integer
            return str(value)
        elif type(value) in [float, np.float64]:  # float
            return "{:.4f}".format(value).replace(".", "p").rstrip("0").rstrip("p")
        elif isinstance(value, list):
            return "_".join(clean_value(v) for v in value)
        elif isinstance(value, tuple):
            return "_".join(clean_value(v) for v in value)
        elif isinstance(value, dict):
            return dict2name(**value)
        elif isinstance(value, Device):
            return clean_name(value.name)
        elif callable(value):
            return value.__name__
        else:
            return clean_name(str(value))
    except TypeError:  # use the __str__ method
        return clean_name(str(value))


class _Dummy:
    pass


@autoname
def _dummy(length=3, wg_width=0.5):
    c = _Dummy()
    c.name = ""
    c.settings = {}
    return c


def test_autoname():
    name_base = _dummy().name
    assert name_base == "_dummy"
    name_int = _dummy(length=3).name
    assert name_int == "_dummy_L3"
    name_float = _dummy(wg_width=0.5).name
    assert name_float == "_dummy_WW0p5"


def test_clean_value():
    assert clean_value(0.5) == "0p5"
    assert clean_value(5) == "5"


def test_clean_name():
    assert clean_name("wg(:_=_2852") == "wg___2852"


if __name__ == "__main__":
    test_autoname()
    # import pp

    # print(clean_value(pp.c.waveguide))

    # c = pp.c.waveguide(polarization="TMeraer")
    # print(c.get_settings()["polarization"])

    # c = pp.c.waveguide(length=11)
    # print(c)
    # pp.show(c)

    # print(clean_name("Waveguidenol1_(:_=_2852"))
    # print(clean_value(1.2))
    # print(clean_value(0.2))
    # print(clean_value([1, [2.4324324, 3]]))
    # print(clean_value([1, 2.4324324, 3]))
    # print(clean_value((0.001, 24)))
    # print(clean_value({"a": 1, "b": 2}))



from pp.add_labels import get_input_label


def get_input_labels(
    io_gratings, ordered_ports, component_name, layer_label, gc_port_name
):
    elements = []
    for i, g in enumerate(io_gratings):
        label = get_input_label(
            port=ordered_ports[i],
            gc=g,
            gc_index=i,
            component_name=component_name,
            layer_label=layer_label,
            gc_port_name=gc_port_name,
        )
        elements += [label]

    return elements



from pp.components.coupler_symmetric import coupler_symmetric
from pp.components.coupler_symmetric import coupler_symmetric_biased
from pp.components.coupler_straight import coupler_straight
from pp.components.coupler_straight import coupler_straight_biased
from pp.netlist_to_gds import netlist_to_component
from pp.name import autoname
from pp.drc import assert_on_2nm_grid
from pp.drc import assert_on_1nm_grid
from pp.add_pins import add_pins


@autoname
def coupler(**kwargs):
    """ symmetric coupler

    Args:
        gap
        length
        coupler_symmetric_factory
        coupler_straight

    .. plot::
      :include-source:

      import pp

      c = pp.c.coupler(gap=0.2, length=10)
      pp.plotgds(c)

    """
    components, connections, ports_map = coupler_netlist(**kwargs)
    component = netlist_to_component(components, connections, ports_map)
    add_pins(component)
    return component


@autoname
def coupler_biased(**kwargs):
    """ symmetric coupler
    """
    components, connections, ports_map = coupler_netlist(
        coupler_symmetric_factory=coupler_symmetric_biased,
        coupler_straight=coupler_straight_biased,
        **kwargs
    )
    component = netlist_to_component(components, connections, ports_map)
    return component


def coupler_netlist(
    wg_width=0.5,
    gap=0.236,
    length=20.007,
    coupler_symmetric_factory=coupler_symmetric,
    coupler_straight=coupler_straight,
):
    """
     SBEND_L-CS-SBEND_R
    """

    assert_on_1nm_grid(length)
    assert_on_2nm_grid(gap)

    _sbend = coupler_symmetric_factory(gap=gap, wg_width=wg_width)
    _cpl_straight = coupler_straight(length=length, gap=gap, width=wg_width)

    components = {
        "SBEND_L": (_sbend, "mirror_y"),
        "SBEND_R": (_sbend, "None"),
        "CS": (_cpl_straight, "None"),
    }

    connections = [("SBEND_L", "W0", "CS", "W0"), ("CS", "E0", "SBEND_R", "W0")]

    ports_map = {
        "W0": ("SBEND_L", "E0"),
        "W1": ("SBEND_L", "E1"),
        "E0": ("SBEND_R", "E0"),
        "E1": ("SBEND_R", "E1"),
    }

    return components, connections, ports_map


if __name__ == "__main__":
    import pp

    # from pp.routing import add_io_optical

    # c = coupler(gap=0.245, length=5.67, wg_width=0.2)
    # c = coupler(gap=0.2, length=5, wg_width=0.4)
    # c = coupler_biased(gap=0.2, length=5, wg_width=0.5)
    c = coupler()
    print(c.get_settings())
    # cc = add_io_optical(c)
    pp.show(c)

    # from pp.routing.connect_component import add_io_optical
    # cc = add_io_optical(c)
    # pp.show(cc)



import functools
import collections
from inspect import signature
from pp.component import NAME_TO_DEVICE
from pp.name import get_component_name


def update_dicts_recurse(target_dict, default_dict):
    target_dict = target_dict.copy()
    default_dict = default_dict.copy()
    for k, v in default_dict.items():
        if k not in target_dict:
            target_dict[k] = v
        else:
            vtype = type(target_dict[k])
            if vtype == dict or vtype == collections.OrderedDict:
                target_dict[k] = update_dicts_recurse(target_dict[k], default_dict[k])
    return target_dict


def autoname_and_cache(component_function):
    """ decorator for auto-naming component functions

    you can pass a name argument
    otherwise it creates a name by concenating all Keyword arguments

    requires kwargs

    .. code-block:: python
        import pp

        @pp.autoname
        def rectangle(widht=2, height=3):
            c = Component()
            return c
    """

    @functools.wraps(component_function)
    def wrapper(*args, **kwargs):
        if args:
            raise ValueError("autoname supports only Keyword args")
        if "name" in kwargs:
            name = kwargs.pop("name")
        else:
            name = get_component_name(component_function.__name__, **kwargs)

        if name in NAME_TO_DEVICE:
            component = NAME_TO_DEVICE[name]
        else:
            component = component_function(**kwargs)
            component.name = name
            component.module = component_function.__module__
            component.function_name = component_function.__name__
            sig = signature(component_function)

            update_dicts_recurse(component.settings, kwargs)
            update_dicts_recurse(
                component.settings, {p.name: p.default for p in sig.parameters.values()}
            )

            # component.settings.update(
            # **{p.name: p.default for p in sig.parameters.values()}
            # )
            # component.settings.update(**kwargs)

            NAME_TO_DEVICE[name] = component

        return component

    return wrapper



import pp
from pp.layers import LAYER
from pp.ports.port_naming import deco_rename_ports
from pp.components.hline import hline

__version__ = "0.0.1"

WIRE_WIDTH = 10.0


@deco_rename_ports
@pp.autoname
def wire(length=50, width=WIRE_WIDTH, layer=LAYER.M3):
    """
    straight wire
    """
    return hline(length=length, width=width, layer=layer)


@deco_rename_ports
@pp.autoname
def corner(width=WIRE_WIDTH, radius=None, layer=LAYER.M3):
    """
    radius passed for consistency with other types of bends
    """
    component = pp.Component()
    a = width / 2
    xpts = [-a, a, a, -a]
    ypts = [-a, -a, a, a]

    component.add_polygon([xpts, ypts], layer=layer)
    component.add_port(name="W0", midpoint=(-a, 0), width=width, orientation=180)
    component.add_port(name="N0", midpoint=(0, a), width=width, orientation=90)
    component.info["length"] = width

    return component


def test_wire():
    c = wire()
    pp.write_gds(c)
    return c


if __name__ == "__main__":
    test_wire()



import pp

from waveguide import waveguide
from bend_circular import bend_circular
from y_splitter import y_splitter


@pp.autoname
def mzi(delta_length=100):
    c = pp.c.mzi(
        L1=delta_length,
        straight_factory=waveguide,
        bend90_factory=bend_circular,
        coupler_factory=y_splitter,
    )
    return c


if __name__ == "__main__":
    c = mzi(delta_length=100)
    pp.show(c)



"""
Route optical and electrical waveguides
"""

from pp.routing.connect_component import add_io_optical
from pp.routing.route_fiber_single import route_fiber_single
from pp.routing.connect_bundle import connect_bundle
from pp.routing.connect_bundle import link_optical_ports
from pp.routing.connect_bundle import link_electrical_ports
from pp.routing.route_ports_to_side import route_ports_to_side
from pp.routing.route_ports_to_side import route_elec_ports_to_side
from pp.routing.manhattan import round_corners
from pp.routing.manhattan import route_manhattan

__all__ = [
    "add_io_optical",
    "route_fiber_single",
    "link_optical_ports",
    "link_electrical_ports",
    "route_ports_to_side",
    "route_elec_ports_to_side",
    "connect_bundle",
    "round_corners",
    "route_manhattan",
]

if __name__ == "__main__":
    print(__all__)



"""autoplacer - placing gds components"""


from pp.autoplacer.auto_placer import AutoPlacer
from pp.autoplacer.library import Library
from pp.autoplacer.chip_array import ChipArray
from pp.autoplacer.cell_list import CellList
from pp.autoplacer.yaml_placer import place_from_yaml

__all__ = ["AutoPlacer", "Library", "CellList", "ChipArray", "place_from_yaml"]


if __name__ == "__main__":
    from pp import klive

    lib = Library()
    mask = ChipArray("chip_array", 25e6, 25e6, 3, 4, lib)
    mask.pack_grid(lib.pop("align"))
    mask.pack_grid(lib.pop(".*"))
    mask.write("chip_array.gds")
    klive.show("chip_array.gds")



"""
write regression tests for component ports and properties
"""

from pp.samples.pdk import component_type2factory

_skip_test = set()
_skip_test_ports = set()


def write_test_properties():
    """ writes a regression test for all the component properties dict"""
    with open("test_components.py", "w") as f:
        f.write(
            "# this code has been automatically generated from pp/samples/pdk/write_tests.py\n"
        )
        f.write("import pp.samples.pdk as pdk\n\n")

        for c in set(component_type2factory.keys()) - _skip_test:
            f.write(
                f"""
def test_{c}(data_regression):
    c = pdk.{c}()
    data_regression.check(c.get_settings())

"""
            )


def write_test_ports():
    """ writes a regression test for all the ports """
    with open("test_ports.py", "w") as f:
        f.write(
            "# this code has been automatically generated from pp/samples/pdk/write_tests.py\n"
        )
        f.write("import pp.samples.pdk as pdk\n\n")

        for component_function in (
            set(component_type2factory.values()) - _skip_test_ports
        ):
            c = component_function.__name__
            if component_function().ports:

                f.write(
                    f"""
def test_{c}(num_regression):
    c = pdk.{c}()
    num_regression.check(c.get_ports_array())

    """
                )


if __name__ == "__main__":
    write_test_properties()
    write_test_ports()



from pp.name import autoname
from pp.layers import LAYER
from pp.components.compass import compass
import pp

WIRE_WIDTH = 10.0


@autoname
def pad(width=100.0, height=100.0, layer=LAYER.M3):
    """ rectangular pad
    """
    c = pp.Component()
    _c = compass(size=(width, height), layer=layer).ref()
    c.add(_c)
    c.absorb(_c)
    c.ports = _c.ports
    return c


@autoname
def pad_array(pad=pad, start=(0, 0), spacing=(150, 0), n=6, port_list=["N"]):
    """ array of rectangular pads
    """
    c = pp.Component()

    for i in range(n):
        p = c << pp.call_if_func(pad)
        p.x = i * spacing[0]
        for port_name in port_list:
            port_name_new = "{}{}".format(port_name, i)
            c.add_port(port=p.ports[port_name], name=port_name_new)
    return c


if __name__ == "__main__":
    # c = pad()
    # c = pad(width=10, height=10)
    # print(c.ports.keys())
    # print(c.settings['spacing'])
    c = pad_array()
    pp.show(c)



# this code has been automatically generated from pp/components/__init__.py
import pp


def test_ring_double_bus(num_regression):
    c = pp.c.ring_double_bus()
    num_regression.check(c.get_ports_array())


def test_bend_euler180(num_regression):
    c = pp.c.bend_euler180()
    num_regression.check(c.get_ports_array())


def test_grating_coupler_elliptical_te(num_regression):
    c = pp.c.grating_coupler_elliptical_te()
    num_regression.check(c.get_ports_array())


def test_coupler(num_regression):
    c = pp.c.coupler()
    num_regression.check(c.get_ports_array())


def test_coupler_symmetric(num_regression):
    c = pp.c.coupler_symmetric()
    num_regression.check(c.get_ports_array())


def test_bend_circular_heater(num_regression):
    c = pp.c.bend_circular_heater()
    num_regression.check(c.get_ports_array())


def test_bend_circular180(num_regression):
    c = pp.c.bend_circular180()
    num_regression.check(c.get_ports_array())


def test_mmi2x2(num_regression):
    c = pp.c.mmi2x2()
    num_regression.check(c.get_ports_array())


def test_mzi1x2(num_regression):
    c = pp.c.mzi1x2()
    num_regression.check(c.get_ports_array())


def test_pad(num_regression):
    c = pp.c.pad()
    num_regression.check(c.get_ports_array())


def test_waveguide_heater(num_regression):
    c = pp.c.waveguide_heater()
    num_regression.check(c.get_ports_array())


def test_coupler_asymmetric(num_regression):
    c = pp.c.coupler_asymmetric()
    num_regression.check(c.get_ports_array())


def test_grating_coupler_uniform(num_regression):
    c = pp.c.grating_coupler_uniform()
    num_regression.check(c.get_ports_array())


def test_coupler90(num_regression):
    c = pp.c.coupler90()
    num_regression.check(c.get_ports_array())


def test_coupler_straight(num_regression):
    c = pp.c.coupler_straight()
    num_regression.check(c.get_ports_array())


def test_spiral_inner_io(num_regression):
    c = pp.c.spiral_inner_io()
    num_regression.check(c.get_ports_array())


def test_spiral_external_io(num_regression):
    c = pp.c.spiral_external_io()
    num_regression.check(c.get_ports_array())


def test_mmi1x2(num_regression):
    c = pp.c.mmi1x2()
    num_regression.check(c.get_ports_array())


def test_spiral_inner_io_euler(num_regression):
    c = pp.c.spiral_inner_io_euler()
    num_regression.check(c.get_ports_array())


def test_taper(num_regression):
    c = pp.c.taper()
    num_regression.check(c.get_ports_array())


def test_waveguide(num_regression):
    c = pp.c.waveguide()
    num_regression.check(c.get_ports_array())


def test_bend_circular(num_regression):
    c = pp.c.bend_circular()
    num_regression.check(c.get_ports_array())


def test_loop_mirror(num_regression):
    c = pp.c.loop_mirror()
    num_regression.check(c.get_ports_array())


def test_wg_heater_connected(num_regression):
    c = pp.c.wg_heater_connected()
    num_regression.check(c.get_ports_array())


def test_compass(num_regression):
    c = pp.c.compass()
    num_regression.check(c.get_ports_array())


def test_bend_s(num_regression):
    c = pp.c.bend_s()
    num_regression.check(c.get_ports_array())


def test_grating_coupler_elliptical_tm(num_regression):
    c = pp.c.grating_coupler_elliptical_tm()
    num_regression.check(c.get_ports_array())


def test_pad_array(num_regression):
    c = pp.c.pad_array()
    num_regression.check(c.get_ports_array())


def test_bend_euler90(num_regression):
    c = pp.c.bend_euler90()
    num_regression.check(c.get_ports_array())


def test_ring_single_bus(num_regression):
    c = pp.c.ring_single_bus()
    num_regression.check(c.get_ports_array())


def test_mzi2x2(num_regression):
    c = pp.c.mzi2x2()
    num_regression.check(c.get_ports_array())



import pp

from pp.ports.port_naming import deco_rename_ports, rename_ports_by_orientation
from pp.components import bend_circular
from pp.components import waveguide
from pp.components import mmi1x2


@deco_rename_ports
@pp.autoname
def mzi(
    L0=1,
    L1=9,
    L2=10,
    bend_radius=10.0,
    bend90_factory=bend_circular,
    straight_factory=waveguide,
    coupler_factory=mmi1x2,
    combiner_factory=None,
):
    """ Mzi adapted for using different coupler and combiner factories

    Args:
        L0: vertical length for both and top arms
        L1: bottom arm extra length
        L2: L_top horizontal length
        bend_radius: 10.0
        bend90_factory: bend_circular
        straight_factory: waveguide
        coupler_factory: coupler
        combiner_factory: coupler

    .. code::

               __L2__
              |      |
              L0     L0r
              |      |
     coupler==|      |==combiner
              |      |
              L0     L0r
              |      |
              L1     L1
              |      |
              |__L2__|


    .. plot::
      :include-source:

      import pp

      c = pp.c.mzi(L0=0.1, L1=0, L2=10)
      pp.plotgds(c)

    """
    c = pp.Component()
    combiner_factory = combiner_factory or coupler_factory

    coupler = pp.call_if_func(coupler_factory)
    combiner = pp.call_if_func(combiner_factory)

    b90 = bend90_factory(radius=bend_radius)
    l0 = straight_factory(length=L0)

    coupler = rename_ports_by_orientation(coupler)
    combiner = rename_ports_by_orientation(combiner)

    y1l = coupler.ports["E0"].y
    y1r = combiner.ports["E0"].y

    y2l = coupler.ports["E1"].y
    y2r = combiner.ports["E1"].y

    dl = abs(y2l - y1l)  # splitter ports distance
    dr = abs(y2r - y1r)  # combiner ports distance
    delta_length = dl - dr
    assert (
        delta_length + L0 > 0
    ), f"input and output couplers height  offset (delta_length)  {delta_length} + {L0} >0"

    l0r = straight_factory(length=L0 + delta_length / 2)

    l1 = straight_factory(length=L1)
    l2 = straight_factory(length=L2)

    cin = c << coupler
    cout = c << combiner

    # top arm
    blt = c << b90
    bltl = c << b90
    bltr = c << b90
    blmr = c << b90  # bend left medium right

    l0tl = c << l0
    l2t = c << l2
    l0tr = c << l0r

    blt.connect(port="W0", destination=cin.ports["E1"])
    l0tl.connect(port="W0", destination=blt.ports["N0"])
    bltl.connect(port="N0", destination=l0tl.ports["E0"])
    l2t.connect(port="W0", destination=bltl.ports["W0"])
    bltr.connect(port="N0", destination=l2t.ports["E0"])
    l0tr.connect(port="W0", destination=bltr.ports["W0"])
    blmr.connect(port="W0", destination=l0tr.ports["E0"])
    cout.connect(port="E0", destination=blmr.ports["N0"])

    # bot arm
    blb = c << b90
    l0bl = c << l0
    l1l = c << l1
    blbl = c << b90
    l2t = c << l2
    brbr = c << b90
    l1r = c << l1
    l0br = c << l0r
    blbmrb = c << b90  # bend left medium right bottom

    blb.connect(port="N0", destination=cin.ports["E0"])
    l0bl.connect(port="W0", destination=blb.ports["W0"])
    l1l.connect(port="W0", destination=l0bl.ports["E0"])
    blbl.connect(port="W0", destination=l1l.ports["E0"])
    l2t.connect(port="W0", destination=blbl.ports["N0"])
    brbr.connect(port="W0", destination=l2t.ports["E0"])

    l1r.connect(port="W0", destination=brbr.ports["N0"])
    l0br.connect(port="W0", destination=l1r.ports["E0"])
    blbmrb.connect(port="N0", destination=l0br.ports["E0"])

    # west ports
    for port_name, port in cin.ports.items():
        if port.angle == 180:
            c.add_port(name=port_name, port=port)

    # east ports
    i = 0
    for port_name, port in cout.ports.items():
        if port.angle == 0:
            c.add_port(name=f"E{i}", port=port)
            i += 1
    return c


if __name__ == "__main__":
    c = mzi()
    # print(c.ports)
    pp.show(c)
    # print(c.get_settings())



from waveguide import waveguide
from bend_circular import bend_circular
from y_splitter import y_splitter
from mzi import mzi


component_type2factory = dict(
    waveguide=waveguide, bend_circular=bend_circular, y_splitter=y_splitter, mzi=mzi
)


__all__ = list(component_type2factory.keys())



import numpy as np

import pp
from pp.components.bezier import bezier


__version__ = "0.0.2"


@pp.autoname
def bend_s(
    width=0.5, height=2, length=10, layer=pp.LAYER.WG, nb_points=99, with_pins=False
):
    """ S bend
    Based on bezier curve

    Args:
        width
        height: in y direction
        length: in x direction
        layer: gds number
        nb_points: number of points

    .. plot::
      :include-source:

      import pp

      c = pp.c.bend_s(width=0.5, height=20)
      pp.plotgds(c)

    """
    l, h = length, height
    c = bezier(
        width=width,
        control_points=[(0, 0), (l / 2, 0), (l / 2, h), (l, h)],
        t=np.linspace(0, 1, nb_points),
        layer=layer,
        with_pins=with_pins,
    )
    c.add_port(name="W0", port=c.ports.pop("0"))
    c.add_port(name="E0", port=c.ports.pop("1"))

    # c.ports["W0"] = c.ports.pop("0")
    # c.ports["E0"] = c.ports.pop("1")
    return c


@pp.ports.port_naming.deco_rename_ports
@pp.autoname
def bend_s_biased(width=0.5, height=2, length=10, layer=pp.LAYER.WG, nb_points=99):
    l, h = length, height
    return bezier(
        width=pp.bias.width(width),
        control_points=[(0, 0), (l / 2, 0), (l / 2, h), (l, h)],
        t=np.linspace(0, 1, nb_points),
        layer=layer,
    )


def _demo():
    c = bend_s()
    pp.write_gds(c)
    return c


if __name__ == "__main__":
    c = bend_s(with_pins=False)
    # c = bend_s_biased()
    # print(c.info["min_bend_radius"])
    pp.show(c)



"""
A port is defined by:

 - position
 - orientation
 - name (which has to be unique within the component being defined)
 - width
 - layer
 - type (optical, dc, rf, detector)

"""


from pp.ports.read_port_markers import read_port_markers
from pp.ports.utils import flipped
from pp.ports.utils import get_ports_facing
from pp.ports.utils import get_non_optical_ports
from pp.ports.utils import get_optical_ports
from pp.ports.utils import select_optical_ports
from pp.ports.utils import select_electrical_ports

from pp.ports.port_naming import deco_rename_ports
from pp.ports.port_naming import auto_rename_ports


__all__ = [
    "read_port_markers",
    "flipped",
    "deco_rename_ports",
    "auto_rename_ports",
    "get_ports_facing",
    "get_optical_ports",
    "get_non_optical_ports",
    "select_optical_ports",
    "select_electrical_ports",
]



"""
This module provides a fairly sophisticated logging system as an alternative to ``print``. You can use ``LOGGER.info(text)`` just like ``print``. However, there are some real advantages.

Rationale
----------

Firstly, ``LOGGER.info()`` doesn't just print to the terminal. It also writes more detailed logs to a file, usually in ``build/logs/main.log``. This gives a permanent record of warnings and errors, which is also synced from Jenkins into Dropbox.  While output on the terminal is simple::

    >>> from pp.logger import LOGGER
    >>> LOGGER.info("Hello world")
    Hello world

The output in the file is more sophisticated:

.. code-block:: none

    2017-11-08 04:05:08,481 [INFO ] [<ipython-input-2-9487df49f71a>:<module>:1] Hello world

Let's break down the structure of this log entry:

=============================================  =================================================================================================================================================================
Part                                           Meaning
=============================================  =================================================================================================================================================================
2017-11-08 04:05:08,481                        A timestamp
[INFO]                                         The log-level, it can be any of ``CRITICAL``, ``ERROR``, ``WARNING``, ``INFO``, ``DEBUG``, ``NOTSET``.
<ipython-input-2-9487df49f71a>:<module>:1      The module, function and line number which emitted this log entry. Really useful when somebody has added annoying terminal output to a complicated codebase.
Hello world                                    The text message
=============================================  =================================================================================================================================================================

Everyday usage
---------------

In 99% of cases all you need to do is this::

    from pp import CONFIG, LOGGER

    ... code ...

    LOGGER.info("Hello world")
    LOGGER.warning("Something bad is happening")

If you want to emit an error or a warning, call ``LOGGER.error("Some text")`` or ``LOGGER.warning("Some text")``. If you want something to be logged but not printed to the terminal, use ``LOGGER.debug``.

gdsdactory stores separate log files for each mask. You can easily review the logs just by reading the files, but you can also use ``pf log show``::

    $ pf log show

It can also post logs to Slack, which can be useful when running on Jenkins or a remote server. Just add a block like this to the local ``~/.gdsfactory/config.yml``:

.. code-block:: yaml

    slack:
      enabled: true
      channel: '<channel>'
      hook: 'https://hooks.slack.com/services/<secret>'

"""


import traceback
import logging
import sys
from os.path import join
from pp.config import CONFIG

if CONFIG.get("slack"):
    import slack


ERROR_COLOR = "#ff0000"  # color name is built in to Slack API
WARNING_COLOR = "warning"  # color name is built in to Slack API
INFO_COLOR = "#439FE0"
DEFAULT_EMOJI = ":godmode:"

COLORS = {
    logging.CRITICAL: ERROR_COLOR,
    logging.FATAL: ERROR_COLOR,
    logging.ERROR: ERROR_COLOR,
    logging.WARNING: WARNING_COLOR,
    logging.INFO: INFO_COLOR,
    logging.DEBUG: INFO_COLOR,
    logging.NOTSET: INFO_COLOR,
}


class SlackHandler(logging.Handler):
    """ A handler which tries to post to Slack """

    def __init__(self):
        logging.Handler.__init__(self)

    def emit(self, record):
        """ Emit a message """
        message = str(record.getMessage())
        if hasattr(record, "raw"):  # The raw flag was set, we need to handle specially
            fallback = message[:10] + "..."
            slack.post_log(
                message="Trace:", data=message, fallback=fallback, level=record.levelno
            )
        elif record.exc_info:  # There is a stack trace, let's show it
            exc = "\n".join(traceback.format_exception(*record.exc_info))
            fallback = str(type(record.exc_info[0]))
            slack.post_log(
                message=message, data=exc, fallback=fallback, level=record.levelno
            )
        else:  # It's just an INFO or something
            slack.post_log(message=message, level=record.levelno)


def get_filehandler(name, level):
    """ Get a filehandler """
    logdir = CONFIG["log_directory"]
    fh = logging.FileHandler(join(logdir, name + ".log"))
    fh.setFormatter(
        logging.Formatter(
            "%(asctime)s [%(levelname)-5.5s] [%(module)s:%(funcName)s:%(lineno)d] %(message)s"
        )
    )
    fh.setLevel(level)
    return fh


def get_streamhandler(level=logging.INFO):
    """ Get a streamhandler """
    sh = logging.StreamHandler()
    sh.setFormatter(logging.Formatter("%(message)s"))
    sh.setLevel(level)
    return sh


def get_slackhandler(level=logging.INFO):
    """ Get a Slack handler """
    sh = SlackHandler()
    sh.setLevel(level)
    return sh


def get_logger(name):
    """ Get a complicated logger """
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # Add a file handler for all messages (including debug messages)
    logger.addHandler(get_filehandler("debug", logging.DEBUG))
    logger.addHandler(get_filehandler(name, logging.INFO))
    logger.addHandler(get_filehandler("error", logging.ERROR))
    logger.addHandler(get_filehandler("failed_to_pack", logging.WARNING))

    # Add a console handler
    logger.addHandler(get_streamhandler())

    # Try to add a Slack handler
    if CONFIG.get("slack"):
        logger.addHandler(get_slackhandler())
    return logger


def handle_exception(exc_type, exc_value, exc_traceback):
    """ This lets us catch errors just before Python dies """
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    LOGGER.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))


if not __name__ == "__main__":
    LOGGER = get_logger("main")
    sys.excepthook = handle_exception
else:
    LOGGER2 = get_logger("test")
    LOGGER2.debug("Nothing")
    LOGGER2.info("Something")
    LOGGER2.error("Simple error")

    LOGGER2.info("I'm about to show some code:")
    LOGGER2.info("here\nis\nsome\ncode", extra={"raw": True})

    LOGGER2.error("Zero div error", extra={"raw": True})

    try:
        1 / 0
    except ZeroDivisionError as e:
        LOGGER2.error("Zero div error", exc_info=e)

    LOGGER2.warning("Simple warning")



import os
import json

import hiyapyco
from pp.mask.parse_xlsx_device_manifest import parse_device_manifest


def parse_mask_json(mask_json_path):
    with open(mask_json_path) as f:
        data = json.loads(f.read())

    ordered_fields = ["name", "json_version", "width", "height"]
    ordered_fields += ["git_hash_autoplacer", "git_hash_mask", "git_hash_pp"]
    ordered_fields += ["description", "does", "cells"]

    return {k: data[k] for k in ordered_fields if k in data}


def parse_csv_data(csv_labels_path):
    with open(csv_labels_path) as f:
        lines = [line.replace("\n", "").split(",") for line in f.readlines()]
        lines = [[s.strip() for s in splitted if s.strip()] for splitted in lines]
        lines = [l for l in lines if l]
    return lines


def get_cell_from_label(label):
    return label.split("(")[1].split(")")[0]


def merge_all_metadata(gdspath, device_manifest_path):
    """ from a gds mask combines test_protocols and labels positions for each DOE
    Do a map cell: does
    Usually each cell will have only one DOE. But in general it should be allowed for a cell to belong to multiple DOEs
    """
    mask_json_path = gdspath.replace(".gds", ".json")
    csv_labels_path = gdspath.replace(".gds", ".csv")
    output_tm_path = gdspath.replace(".gds", ".tp.json")
    tm_dict = {}

    device_manifest_data = parse_device_manifest(device_manifest_path)

    mask_directory = os.path.split(gdspath)[0]
    mask_build_directory = os.path.split(mask_directory)[0]
    mask_root_directory = os.path.split(mask_build_directory)[0]
    test_protocols_path = os.path.join(mask_root_directory, "test_protocols.yml")
    analysis_protocols_path = os.path.join(
        mask_root_directory, "data_analysis_protocols.yml"
    )

    assert os.path.isfile(mask_json_path), "missing {}".format(mask_json_path)
    assert os.path.isfile(csv_labels_path), "missing {}".format(csv_labels_path)

    mask_data = parse_mask_json(mask_json_path)

    if os.path.isfile(test_protocols_path):
        test_protocols = hiyapyco.load(test_protocols_path)
        tm_dict["test_protocols"] = test_protocols
    if os.path.isfile(analysis_protocols_path):
        analysis_protocols = hiyapyco.load(analysis_protocols_path)
        tm_dict["analysis_protocols"] = analysis_protocols

    data = parse_csv_data(csv_labels_path)
    # cell_x_y = [(get_cell_from_label(l), x, y) for l, x, y in data]

    does = mask_data["does"]
    cells = mask_data["cells"]
    doe_devices = set()
    for doe in does.values():
        doe_devices.update(doe["cells"])

    # from pprint import pprint
    # pprint(list(cells.keys()))
    # pprint(list(device_manifest_data.keys()))

    ## Inject device manifest data into cells
    for cell_name in doe_devices:

        if cell_name not in cells:
            print(
                "skip reconcile data for cell {} - cell not in cells".format(cell_name)
            )
            continue

        if cell_name not in device_manifest_data:
            print(
                "skip reconcile data for cell {} - cell not in device manifest".format(
                    cell_name
                )
            )
            continue
        cells[cell_name].update(device_manifest_data[cell_name])

    ## Map cell to DOEs - generic case where a cell could belong to multiple experiments
    cell_to_does = {}
    for doe_name, doe in does.items():
        for c in doe["cells"]:
            if c not in cell_to_does:
                cell_to_does[c] = set()
            cell_to_does[c].update([doe_name])

    tm_dict["does"] = {}
    doe_tm = tm_dict["does"]
    doe_tm.update(does)
    for doe_name, doe in doe_tm.items():
        doe.pop("cells")
        doe["instances"] = {}

    ## Cell instances which need to be measured MUST have a unique cell name
    for label, x, y in data:
        cell_name = get_cell_from_label(label)
        if cell_name not in cell_to_does:
            continue
        cell_does = cell_to_does[cell_name]
        for doe_name in cell_does:
            _doe = doe_tm[doe_name]

            if cell_name not in _doe["instances"]:
                # Unique Cell instance to labels and coordinates
                _doe["instances"][cell_name] = []
            _doe["instances"][cell_name].append({"label": label, "x": x, "y": y})

    # Adding the cells settings
    tm_dict["cells"] = cells

    with open(output_tm_path, "w") as json_out:
        json.dump(tm_dict, json_out, indent=2)


if __name__ == "__main__":
    import pp

    gdspath = os.path.join(
        pp.CONFIG["masks_path"], "sample", "build", "mask", "sample.gds"
    )
    merge_all_metadata(gdspath)



from numpy import pi
import pp
from pp.components import taper
from pp.components import bend_circular
from pp.components import waveguide
from pp.routing.manhattan import round_corners
from pp.config import WG_EXPANDED_WIDTH, TAPER_LENGTH


@pp.autoname
def delay_snake(
    total_length=160000,
    L0=2350.0,
    n=5,
    taper=taper,
    bend_factory=bend_circular,
    bend_radius=10.0,
    straight_factory=waveguide,
    wg_width=0.5,
):
    """ Snake input facing west
    Snake output facing east

    Args:
        total_length:
        L0:
        n:
        taper:
        bend_factory
        bend_radius
        straight_factory
        wg_width

    .. code::

       | L0 |    L2        |

            ->-------------|
                           | pi * radius
       |-------------------|
       |
       |------------------->

       |        L1         |

    .. plot::
      :include-source:

      import pp

      c = pp.c.delay_snake(L0=5, total_length=1600, n=2)
      pp.plotgds(c)

    """
    epsilon = 0.1
    R = bend_radius
    bend90 = bend_factory(radius=R, width=wg_width)
    L1 = (total_length + L0 - n * (pi * R + epsilon)) / (2 * n + 1)
    L2 = L1 - L0
    assert (
        L2 > 0
    ), "Snake is too short: either reduce L0, increase the total length,\
    or decrease n"

    y = 0
    path = [(0, y), (L2, y)]
    for i in range(n):
        y -= 2 * R + epsilon
        path += [(L2, y), (-L0, y)]
        y -= 2 * R + epsilon
        path += [(-L0, y), (L2, y)]

    path = [(round(_x, 3), round(_y, 3)) for _x, _y in path]

    component = pp.Component()
    if taper:
        if callable(taper):
            _taper = taper(
                width1=wg_width, width2=WG_EXPANDED_WIDTH, length=TAPER_LENGTH
            )
        else:
            _taper = taper
    snake = round_corners(path, bend90, straight_factory, taper=_taper)
    component.add(snake)
    component.ports = snake.ports

    pp.ports.port_naming.auto_rename_ports(component)
    return component


if __name__ == "__main__":
    c = delay_snake()
    pp.show(c)



import functools


def deco_rename_ports(component_factory):
    @functools.wraps(component_factory)
    def auto_named_component_factory(*args, **kwargs):
        device = component_factory(*args, **kwargs)
        auto_rename_ports(device)
        return device

    return auto_named_component_factory


def _rename_ports_facing_side(direction_ports, prefix=""):
    for direction, list_ports in list(direction_ports.items()):

        if direction in ["E", "W"]:
            # first sort along x then y
            list_ports.sort(key=lambda p: p.x)
            list_ports.sort(key=lambda p: p.y)

        if direction in ["S", "N"]:
            # first sort along y then x
            list_ports.sort(key=lambda p: p.y)
            list_ports.sort(key=lambda p: p.x)

        for i, p in enumerate(list_ports):
            lbl = prefix + direction + str(i)
            p.name = lbl


def rename_ports_by_orientation(device, layers_excluded=[]):
    """
    Assign standard port names based on the layer of the port
    """

    # Naming functions

    direction_ports = {x: [] for x in ["E", "N", "W", "S"]}

    ports_on_process = [
        p for p in device.ports.values() if p.layer not in layers_excluded
    ]

    for p in ports_on_process:
        # Make sure we can backtrack the parent component from the port
        p.parent = device

        angle = p.orientation % 360
        if angle <= 45 or angle >= 315:
            direction_ports["E"].append(p)
        elif angle <= 135 and angle >= 45:
            direction_ports["N"].append(p)
        elif angle <= 225 and angle >= 135:
            direction_ports["W"].append(p)
        else:
            direction_ports["S"].append(p)

    _rename_ports_facing_side(direction_ports)
    device.ports = {p.name: p for p in device.ports.values()}
    return device


def auto_rename_ports(device):
    """
    Assign standard port names based on the layer of the port
    """

    def _counter_clockwise(_direction_ports, prefix=""):

        east_ports = _direction_ports["E"]
        east_ports.sort(key=lambda p: p.y)  # sort south to north

        north_ports = _direction_ports["N"]
        north_ports.sort(key=lambda p: -p.x)  # sort east to west

        west_ports = _direction_ports["W"]
        west_ports.sort(key=lambda p: -p.y)  # sort north to south

        south_ports = _direction_ports["S"]
        south_ports.sort(key=lambda p: p.x)  # sort west to east

        ports = east_ports + north_ports + west_ports + south_ports

        for i, p in enumerate(ports):
            p.name = "{}{}".format(prefix, i)

    type_to_ports_naming_functions = {
        "optical": _rename_ports_facing_side,
        "heater": lambda _d: _counter_clockwise(_d, "H_"),
        "dc": lambda _d: _counter_clockwise(_d, "E_"),
        "superconducting": lambda _d: _counter_clockwise(_d, "SC_"),
    }

    type_to_ports = {}

    for p in device.ports.values():
        if p.port_type not in type_to_ports:
            type_to_ports[p.port_type] = []
        type_to_ports[p.port_type] += [p]

    for port_type, port_group in type_to_ports.items():
        if port_type in type_to_ports_naming_functions:
            _func_name_ports = type_to_ports_naming_functions[port_type]
        else:
            raise ValueError(
                "Unknown port type <{}> in device {}, port {}".format(
                    port_type, device.name, p
                )
            )

        # Make sure we can backtrack the parent component from the port

        direction_ports = {x: [] for x in ["E", "N", "W", "S"]}
        for p in port_group:
            p.parent = device
            angle = p.orientation % 360
            if angle <= 45 or angle >= 315:
                direction_ports["E"].append(p)
            elif angle <= 135 and angle >= 45:
                direction_ports["N"].append(p)
            elif angle <= 225 and angle >= 135:
                direction_ports["W"].append(p)
            else:
                direction_ports["S"].append(p)

        _func_name_ports(direction_ports)

    # Set the port dictionnary with the new names
    device.ports = {p.name: p for p in device.ports.values()}
    return device



# this code has been automatically generated from pp/components/__init__.py
import pp


def test_grating_coupler_elliptical_te(data_regression):
    c = pp.c.grating_coupler_elliptical_te()
    data_regression.check(c.get_settings())


def test_rectangle(data_regression):
    c = pp.c.rectangle()
    data_regression.check(c.get_settings())


def test_hline(data_regression):
    c = pp.c.hline()
    data_regression.check(c.get_settings())


def test_bend_euler90(data_regression):
    c = pp.c.bend_euler90()
    data_regression.check(c.get_settings())


def test_ring(data_regression):
    c = pp.c.ring()
    data_regression.check(c.get_settings())


def test_spiral_inner_io_euler(data_regression):
    c = pp.c.spiral_inner_io_euler()
    data_regression.check(c.get_settings())


def test_crossing(data_regression):
    c = pp.c.crossing()
    data_regression.check(c.get_settings())


def test_grating_coupler_te(data_regression):
    c = pp.c.grating_coupler_te()
    data_regression.check(c.get_settings())


def test_litho_calipers(data_regression):
    c = pp.c.litho_calipers()
    data_regression.check(c.get_settings())


def test_mmi2x2(data_regression):
    c = pp.c.mmi2x2()
    data_regression.check(c.get_settings())


def test_test_resistance(data_regression):
    c = pp.c.test_resistance()
    data_regression.check(c.get_settings())


def test_spiral_inner_io(data_regression):
    c = pp.c.spiral_inner_io()
    data_regression.check(c.get_settings())


def test_compass(data_regression):
    c = pp.c.compass()
    data_regression.check(c.get_settings())


def test_coupler_symmetric(data_regression):
    c = pp.c.coupler_symmetric()
    data_regression.check(c.get_settings())


def test_grating_coupler_tm(data_regression):
    c = pp.c.grating_coupler_tm()
    data_regression.check(c.get_settings())


def test_spiral_external_io(data_regression):
    c = pp.c.spiral_external_io()
    data_regression.check(c.get_settings())


def test_coupler(data_regression):
    c = pp.c.coupler()
    data_regression.check(c.get_settings())


def test_spiral_circular(data_regression):
    c = pp.c.spiral_circular()
    data_regression.check(c.get_settings())


def test_grating_coupler_uniform(data_regression):
    c = pp.c.grating_coupler_uniform()
    data_regression.check(c.get_settings())


def test_text(data_regression):
    c = pp.c.text()
    data_regression.check(c.get_settings())


def test_wg_heater_connected(data_regression):
    c = pp.c.wg_heater_connected()
    data_regression.check(c.get_settings())


def test_bend_euler180(data_regression):
    c = pp.c.bend_euler180()
    data_regression.check(c.get_settings())


def test_test_via(data_regression):
    c = pp.c.test_via()
    data_regression.check(c.get_settings())


def test_litho_star(data_regression):
    c = pp.c.litho_star()
    data_regression.check(c.get_settings())


def test_pad(data_regression):
    c = pp.c.pad()
    data_regression.check(c.get_settings())


def test_loop_mirror(data_regression):
    c = pp.c.loop_mirror()
    data_regression.check(c.get_settings())


def test_ring_single_bus(data_regression):
    c = pp.c.ring_single_bus()
    data_regression.check(c.get_settings())


def test_component_lattice(data_regression):
    c = pp.c.component_lattice()
    data_regression.check(c.get_settings())


def test_bend_circular_heater(data_regression):
    c = pp.c.bend_circular_heater()
    data_regression.check(c.get_settings())


def test_delay_snake(data_regression):
    c = pp.c.delay_snake()
    data_regression.check(c.get_settings())


def test_grating_coupler_elliptical_tm(data_regression):
    c = pp.c.grating_coupler_elliptical_tm()
    data_regression.check(c.get_settings())


def test_waveguide_heater(data_regression):
    c = pp.c.waveguide_heater()
    data_regression.check(c.get_settings())


def test_bezier(data_regression):
    c = pp.c.bezier()
    data_regression.check(c.get_settings())


def test_cross(data_regression):
    c = pp.c.cross()
    data_regression.check(c.get_settings())


def test_circle(data_regression):
    c = pp.c.circle()
    data_regression.check(c.get_settings())


def test_pad_array(data_regression):
    c = pp.c.pad_array()
    data_regression.check(c.get_settings())


def test_bend_circular(data_regression):
    c = pp.c.bend_circular()
    data_regression.check(c.get_settings())


def test_rectangle_centered(data_regression):
    c = pp.c.rectangle_centered()
    data_regression.check(c.get_settings())


def test_coupler_asymmetric(data_regression):
    c = pp.c.coupler_asymmetric()
    data_regression.check(c.get_settings())


def test_bend_s(data_regression):
    c = pp.c.bend_s()
    data_regression.check(c.get_settings())


def test_taper(data_regression):
    c = pp.c.taper()
    data_regression.check(c.get_settings())


def test_grating_coupler_tree(data_regression):
    c = pp.c.grating_coupler_tree()
    data_regression.check(c.get_settings())


def test_mzi_arm(data_regression):
    c = pp.c.mzi_arm()
    data_regression.check(c.get_settings())


def test_bend_circular180(data_regression):
    c = pp.c.bend_circular180()
    data_regression.check(c.get_settings())


def test_taper_strip_to_ridge(data_regression):
    c = pp.c.taper_strip_to_ridge()
    data_regression.check(c.get_settings())


def test_waveguide(data_regression):
    c = pp.c.waveguide()
    data_regression.check(c.get_settings())


def test_litho_steps(data_regression):
    c = pp.c.litho_steps()
    data_regression.check(c.get_settings())


def test_coupler_straight(data_regression):
    c = pp.c.coupler_straight()
    data_regression.check(c.get_settings())


def test_mmi1x2(data_regression):
    c = pp.c.mmi1x2()
    data_regression.check(c.get_settings())


def test_ellipse(data_regression):
    c = pp.c.ellipse()
    data_regression.check(c.get_settings())


def test_coupler90(data_regression):
    c = pp.c.coupler90()
    data_regression.check(c.get_settings())


def test_ring_double_bus(data_regression):
    c = pp.c.ring_double_bus()
    data_regression.check(c.get_settings())



import pp
from pp.container import container


@pp.ports.deco_rename_ports
def add_tapers(component, taper, new_component=True):
    """ returns tapered component
    can add tapers over the same component or create a new component"""

    if new_component:
        c = pp.Component(name=component.name + "_t")
        c.add_ref(component)
    else:
        c = component

    for port_name, port in component.ports.copy().items():
        taper_ref = c << pp.call_if_func(taper)
        taper_ref.connect(taper_ref.ports["2"].name, port)
        if not new_component:
            c.ports.pop(port_name)
        c.add_port(name=port_name, port=taper_ref.ports["1"])
    return c


@container
@pp.ports.deco_rename_ports
def add_tapers2(component, taper, suffix="t"):
    """ returns tapers for component """

    c = pp.Component(name=f"{component.name}_{suffix}")

    for port_name, port in component.ports.copy().items():
        taper_ref = c << pp.call_if_func(taper)
        taper_ref.connect(taper_ref.ports["2"].name, port)
        c.add_port(name=port_name, port=taper_ref.ports["1"])
    c.add_ref(component)
    return c


if __name__ == "__main__":
    c = pp.c.waveguide(width=2)
    t = pp.c.taper(width2=2)
    cc = add_tapers2(component=c, taper=t, suffix="t")
    print(cc.ports.keys())
    print(cc.settings.keys())
    pp.show(cc)



import pp
from pp.rotate import rotate
from pp.routing.connect_component import route_fiber_array

from pp.components import waveguide
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_te


def route_fiber_single(
    component,
    optical_io_spacing=50,
    grating_coupler=grating_coupler_te,
    straight_factory=waveguide,
    min_input2output_spacing=230,
    optical_routing_type=2,
    optical_port_labels=None,
    excluded_ports=[],
    **kwargs
):
    """ Returns component I/O for optical testing with single input and oputput fibers (no fiber array)

    Args:
        component: to add grating couplers
        optical_io_spacing: between grating couplers
        grating_coupler:
        straight_factory
        min_input2output_spacing: so opposite fibers do not touch
        optical_routing_type: 0, 1, 2

    .. plot::
      :include-source:

       import pp
       from pp.routing import add_io_optical
       from pp.routing import route_fiber_single

       c = pp.c.mmi1x2()
       cc = add_io_optical(c, get_route_factory=route_fiber_single)
       pp.plotgds(cc)
    """
    component_name = component.name

    if optical_port_labels is None:
        optical_ports = component.get_optical_ports()
    else:
        optical_ports = [component.ports[lbl] for lbl in optical_port_labels]
    optical_ports = [p for p in optical_ports if p.name not in excluded_ports]
    N = len(optical_ports)

    if isinstance(grating_coupler, list):
        grating_couplers = [pp.call_if_func(g) for g in grating_coupler]
        grating_coupler = grating_couplers[0]
    else:
        grating_coupler = pp.call_if_func(grating_coupler)
        grating_couplers = [grating_coupler] * N

    gc_port2center = getattr(grating_coupler, "port2center", grating_coupler.xsize / 2)
    if component.xsize + 2 * gc_port2center < min_input2output_spacing:
        fanout_length = (
            pp.drc.snap_to_5nm_grid(
                min_input2output_spacing - component.xsize - 2 * gc_port2center
            )
            / 2
        )
    else:
        fanout_length = None

    west_ports = [p for p in component.get_optical_ports() if p.name.startswith("W")]
    east_ports = [
        p for p in component.get_optical_ports() if not p.name.startswith("W")
    ]

    # add west input grating couplers
    component.ports = {p.name: p for p in west_ports}
    component = component.rotate(90)

    elements_east, io_grating_lines_east, _ = route_fiber_array(
        component=component,
        with_align_ports=False,
        optical_io_spacing=optical_io_spacing,
        fanout_length=fanout_length,
        grating_coupler=grating_couplers[0],
        optical_routing_type=optical_routing_type,
        **kwargs
    )
    component = rotate(component, angle=-90)

    component.ports = {p.name: p for p in east_ports}
    component = rotate(component, angle=-90)
    component.name = component_name
    elements_west, io_grating_lines_west, _ = route_fiber_array(
        component=component,
        with_align_ports=False,
        optical_io_spacing=optical_io_spacing,
        fanout_length=fanout_length,
        grating_coupler=grating_couplers[1:],
        **kwargs
    )
    for e in elements_west:
        elements_east.append(e.rotate(180))

    for io in io_grating_lines_west[0]:
        io_grating_lines_east.append(io.rotate(180))

    return elements_east, io_grating_lines_east, None


if __name__ == "__main__":
    gcte = pp.c.grating_coupler_te
    gctm = pp.c.grating_coupler_tm

    c = pp.c.mmi2x2()
    # c = pp.c.waveguide()

    elements, gc, _ = route_fiber_single(c, grating_coupler=[gcte, gctm, gcte, gctm])
    for e in elements:
        c.add(e)
    for e in gc:
        c.add(e)
    pp.show(c)



from phidl import device_layout as pd
import pp
from pp.routing.manhattan import round_corners
from pp.rotate import rotate
from pp.add_labels import get_input_label

from import_gds import import_gds
from layers import LAYER
from waveguide import waveguide
from bend_circular import bend_circular


def gc_te1550():
    c = import_gds("ebeam_gc_te1550")
    c = rotate(c, 180)
    c.polarization = "te"
    c.wavelength = 1550
    return c


def gc_te1550_broadband():
    c = import_gds("ebeam_gc_te1550_broadband")
    return c


def gc_te1310():
    c = import_gds("ebeam_gc_te1310")
    c.polarization = "te"
    c.wavelength = 1310
    return c


def gc_tm1550():
    c = import_gds("ebeam_gc_tm1550")
    c.polarization = "tm"
    c.wavelength = 1550
    return c


def connect_strip(
    way_points=[],
    bend_factory=bend_circular,
    straight_factory=waveguide,
    bend_radius=10.0,
    wg_width=0.5,
    **kwargs,
):
    """
    Returns a deep-etched route formed by the given way_points with
    bends instead of corners and optionally tapers in straight sections.
    """
    bend90 = bend_factory(radius=bend_radius, width=wg_width)
    connector = round_corners(way_points, bend90, straight_factory)
    return connector


@pp.autoname
def taper_factory(layer=LAYER.WG, layers_cladding=[], **kwargs):
    c = pp.c.taper(layer=layer, layers_cladding=layers_cladding, **kwargs)
    return c


def get_optical_text(port, gc, gc_index=None, component_name=None):
    polarization = gc.get_property("polarization")
    wavelength_nm = gc.get_property("wavelength")

    assert polarization in [
        "te",
        "tm",
    ], f"Not valid polarization {polarization} in [te, tm]"
    assert (
        isinstance(wavelength_nm, (int, float)) and 1000 < wavelength_nm < 2000
    ), f"{wavelength_nm} is Not valid 1000 < wavelength < 2000"

    if component_name:
        name = component_name

    elif type(port.parent) == pp.Component:
        name = port.parent.name
    else:
        name = port.parent.ref_cell.name

    if isinstance(gc_index, int):
        text = f"opt_in_{polarization}_{int(wavelength_nm)}_({name})_{gc_index}_{port.name}"
    else:
        text = f"opt_in_{polarization}_{int(wavelength_nm)}_({name})_{port.name}"

    return text


gc_port_name = "W0"
layer_label = LAYER.LABEL


def get_input_labels_all(
    io_gratings,
    ordered_ports,
    component_name,
    layer_label=layer_label,
    gc_port_name=gc_port_name,
):
    elements = []
    for i, g in enumerate(io_gratings):
        label = get_input_label(
            port=ordered_ports[i],
            gc=g,
            gc_index=i,
            component_name=component_name,
            layer_label=layer_label,
            gc_port_name=gc_port_name,
        )
        elements += [label]

    return elements


def get_input_labels(
    io_gratings,
    ordered_ports,
    component_name,
    layer_label=layer_label,
    gc_port_name=gc_port_name,
    port_index=1,
):
    if port_index == -1:
        return get_input_labels_all(
            io_gratings=io_gratings,
            ordered_ports=ordered_ports,
            component_name=component_name,
            gc_port_name=gc_port_name,
            port_index=port_index,
        )
    gc = io_gratings[port_index]
    port = ordered_ports[1]

    text = get_optical_text(port=port, gc=gc, component_name=component_name)
    layer, texttype = pd._parse_layer(layer_label)
    label = pd.Label(
        text=text,
        position=gc.ports[gc_port_name].midpoint,
        anchor="o",
        layer=layer,
        texttype=texttype,
    )
    return [label]


def add_gc(
    component,
    layer_label=LAYER.LABEL,
    grating_coupler=gc_te1550,
    bend_factory=bend_circular,
    straight_factory=waveguide,
    taper_factory=taper_factory,
    route_filter=connect_strip,
    gc_port_name="W0",
    get_input_labels_function=get_input_labels,
):
    c = pp.routing.add_io_optical(
        component,
        bend_factory=bend_factory,
        straight_factory=straight_factory,
        route_filter=route_filter,
        grating_coupler=grating_coupler,
        layer_label=layer_label,
        taper_factory=taper_factory,
        gc_port_name=gc_port_name,
        get_input_labels_function=get_input_labels_function,
    )
    c = rotate(c, -90)
    return c


if __name__ == "__main__":
    c = gc_te1550()
    # print(c.ports)
    c = add_gc(component=waveguide())
    pp.show(c)



import pp


def is_electrical_port(port):
    return port.port_type in ["dc", "rf"]


def select_ports(ports, port_type):
    """
    Args:
        ports: a port dictionnary {port name: port} (as returned by Component.ports)
        layers: a list of port layer or a port type (layer or string)

    Returns:
        Dictionnary containing only the ports with the wanted type(s)
        {port name: port}
    """

    # Make it accept Component or ComponentReference
    if isinstance(ports, pp.Component) or isinstance(ports, pp.ComponentReference):
        ports = ports.ports

    return {p_name: p for p_name, p in ports.items() if p.port_type == port_type}


def select_heater_ports(ports):
    return select_ports(ports, port_type="heater")


def select_optical_ports(ports):
    return select_ports(ports, port_type="optical")


def get_optical_ports(ports):
    return select_optical_ports(ports)


def select_electrical_ports(ports):
    d = select_ports(ports, port_type="dc")
    d.update(select_ports(ports, port_type="electrical"))
    return d


def select_dc_ports(ports):
    return select_ports(ports, port_type="dc")


def select_rf_ports(ports):
    return select_ports(ports, port_type="rf")


def select_superconducting_ports(ports):
    d = select_ports(ports, port_type="detector")
    d.update(select_ports(ports, port_type="superconducting"))
    return d


def flipped(port):
    _port = port._copy()
    _port.orientation = (_port.orientation + 180) % 360
    return _port


def move_copy(port, x=0, y=0):
    _port = port._copy()
    _port.midpoint += (x, y)
    return _port


def get_ports_facing(ports, direction="W"):
    if isinstance(ports, dict):
        ports = list(ports.values())
    elif isinstance(ports, pp.Component) or isinstance(ports, pp.ComponentReference):
        ports = list(ports.ports.values())

    direction_ports = {x: [] for x in ["E", "N", "W", "S"]}

    for p in ports:
        angle = p.orientation % 360
        if angle <= 45 or angle >= 315:
            direction_ports["E"].append(p)
        elif angle <= 135 and angle >= 45:
            direction_ports["N"].append(p)
        elif angle <= 225 and angle >= 135:
            direction_ports["W"].append(p)
        else:
            direction_ports["S"].append(p)

    return direction_ports[direction]


def get_non_optical_ports(ports):
    if isinstance(ports, dict):
        ports = list(ports.values())
    elif isinstance(ports, pp.Component) or isinstance(ports, pp.ComponentReference):
        ports = list(ports.ports.values())
    res = [p for p in ports if p.port_type not in ["optical"]]
    return res



"""
Sometimes, when a component is mostly composed of sub-components adjacent to each other, it can be easier to define the component by:

- subcomponents
- connections
- exposed ports in the new components

This can be done using a netlist based approach where these 3 parts are defined:

- components: a dictionnary of {component reference name: (component, transform)}
- connections: a list of (component ref name 1, port name A, component ref name 2, port name B)
- ports_map: a dictionnary of which ports are being exposed together with their new name {port_name: (component ref name, port name)}

The code below illustrates how a simple MZI can be formed using this method.

"""


import pp
from pp.netlist_to_gds import netlist_to_component


@pp.autoname
def test_simple_mzi2x2(
    coupler_length=20.147, arm_length=40, gap=0.234, waveguide_factory=pp.c.waveguide
):
    """
       arm_top
        _____
       |     |
   CP1=       =CP2=
       |_____|

        arm_bot
   """

    # Define the components to use
    coupler = pp.c.coupler(length=coupler_length, gap=gap)
    arm = waveguide_factory(length=arm_length)

    # Create component references and give them unique names {name : (component, transform)}
    # Transform can be either "None", "mirror_x", "mirror_y", "R90", "R180", "R270"

    components = {
        "CP1": (coupler, "None"),
        "CP2": (coupler, "None"),
        "arm_top": (arm, "None"),
        "arm_bot": (arm, "mirror_x"),
    }

    # Provide how components are connected (component 1, port A, component 2, port B)
    # means that port B from component 2 is positioned at port A from component 1

    connections = [
        ## Top arm
        ("CP1", "E1", "arm_top", "W0"),
        ("arm_top", "E0", "CP2", "W1"),
        ## Bottom arm
        ("CP1", "E0", "arm_bot", "W0"),
        ("arm_bot", "E0", "CP2", "W0"),
    ]

    # Create the ports for this component, choosing which ports from the subcomponents should be exposed
    # {port name: (subcomponent, subcomponent port name)}

    ports_map = {
        "W0": ("CP1", "W0"),
        "W1": ("CP1", "W1"),
        "E0": ("CP2", "E0"),
        "E1": ("CP2", "E1"),
    }

    component = netlist_to_component(components, connections, ports_map)
    return component


def test_mzi():
    c = test_simple_mzi2x2()
    assert c
    pp.show(c)


def test_mzi_heated():
    c = test_simple_mzi2x2(waveguide_factory=pp.c.waveguide_heater)
    assert c
    pp.show(c)


if __name__ == "__main__":
    test_mzi_heated()



import pathlib
import pp

from layers import LAYER

cwd = pathlib.Path(__file__).parent.absolute()
gds = cwd / "gds"


layer = LAYER.WG
port_width = 0.5


def position2orientation(p):
    """ we assume that ports with x<0 are inputs (orientation=180deg)
    and ports with x>0 are outputs
    """
    if p[0] < 0:
        return 180
    else:
        return 0


def import_gds(gdsname):
    """ import gds from SIEPIC PDK
    """
    c = pp.import_gds(gds / f"{gdsname}.gds")

    for label in c.get_labels():
        if label.text.startswith("pin"):
            port = pp.Port(
                name=label.text,
                midpoint=label.position,
                width=port_width,
                orientation=position2orientation(label.position),
                layer=layer,
            )
            c.add_port(port)

    return c


if __name__ == "__main__":
    gdsname = "ebeam_y_1550"
    c = import_gds(gdsname)
    pp.show(c)



import pp
from layers import LAYER


@pp.autoname
def waveguide(layer=LAYER.WG, layers_cladding=[], **kwargs):
    c = pp.c.waveguide(layer=layer, layers_cladding=layers_cladding, **kwargs)
    return c


if __name__ == "__main__":
    c = waveguide()
    pp.show(c)



import json
import xlrd
from pp.config import logging


def parse_device_manifest(
    filepath,
    sheet_name_prefix="Device Manifest",
    row_indices=[0, 1, 2, 3],
    key_field_id=2,
):
    """
    Open device manifest file
    Returns {device_id: {attribute: value}}
    """
    book = xlrd.open_workbook(filepath)

    device_manifest_sheet_names = []

    row_indices_attributes = row_indices[:]
    try:
        row_indices_attributes.remove(key_field_id)
    except:
        print(row_indices_attributes)
        raise

    ## Find all device manifest sheets
    device_manifest_sheet_names = [
        s for s in book.sheet_names() if s.startswith(sheet_name_prefix)
    ]

    devices_dict = {}

    index_of_key_field_id = row_indices.index(key_field_id)

    for sheet_name in device_manifest_sheet_names:
        # print(sheet_name)
        device_manifest = book.sheet_by_name(sheet_name)

        rows_gen = device_manifest.get_rows()
        rows = []
        first_row = next(rows_gen)
        col_names = [
            first_row[i].value.replace(" ", "_") for i in row_indices_attributes
        ]
        print(sheet_name)
        print(col_names)
        print()

        for i in range(1000):
            try:
                _row = next(rows_gen)
            except:
                break
            if _row[0].value:
                rows += [[_clean_value(_row[j].value) for j in row_indices]]
                logging.debug(rows[-1][key_field_id])

        _devices_dict = {}
        for row in rows:
            device_id = row.pop(index_of_key_field_id)
            # print(row)
            _devices_dict[device_id] = {_c: _v for _c, _v in zip(col_names, row)}

            devices_dict.update(_devices_dict)
    return devices_dict


def _clean_value(x):
    x = x.strip()
    splitted = x.split(":")
    if len(splitted) == 1:
        return x
    elif len(splitted) == 2:
        return splitted[1]
    else:
        raise ValueError("invalid value {}".format(x))


def inject_device_manifest_data_in_json(
    dm_filepath, json_filepath, dm_parser_settings={}
):

    device_manifest_data = parse_device_manifest(dm_filepath, **dm_parser_settings)

    with open(json_filepath) as f:
        mask_data = json.loads(f.read())

        does = mask_data["does"]
        cells = mask_data["cells"]
        doe_devices = set()
        for doe in does.values():
            doe_devices.update(doe["cells"])

    ## Inject device manifest data into cells
    for cell_name in doe_devices:

        if cell_name not in cells:
            logging.warning(
                "skip reconcile data for cell {} - cell not in cells".format(cell_name)
            )
            continue

        if cell_name not in device_manifest_data:
            logging.warning(
                "skip reconcile data for cell {} - cell not in device manifest".format(
                    cell_name
                )
            )
            continue

        cells[cell_name].update(device_manifest_data[cell_name])

    mask_data["cells"].update(cells)

    with open(json_filepath, "w") as fw:
        fw.write(json.dumps(mask_data, indent=2))


def inject_data_analysis_params_in_json(
    dm_filepath,
    json_filepath,
    dm_parser_settings={
        "sheet_name_prefix": "Analysis - ",
        "row_indices": [4, 6, 7, 8, 9, 10, 11, 12, 13],
        "key_field_id": 4,
    },
    data_analysis_id_from_cell_name=lambda s: s,
):

    device_id_to_analysis_metadata = parse_device_manifest(
        dm_filepath, **dm_parser_settings
    )

    with open(json_filepath) as f:
        mask_data = json.loads(f.read())

        does = mask_data["does"]
        cells = mask_data["cells"]
        doe_devices = set()
        for doe in does.values():
            doe_devices.update(doe["cells"])

    ## Inject device manifest data into cells
    for cell_name in doe_devices:
        cell = cells[cell_name]

        try:
            cell_id = data_analysis_id_from_cell_name(cell_name)
            cell["analysis"] = device_id_to_analysis_metadata[cell_id]
        except:
            print("No match for data analysis on ", cell_name)

    mask_data["cells"].update(cells)

    with open(json_filepath, "w") as fw:
        fw.write(json.dumps(mask_data, indent=2))



import pp


@pp.autoname
def cross(length=10, width=3, layer=0):
    """Generates a right-angle cross from two rectangles of specified length and width.

    Args:
        length: float Length of the cross from one end to the other
        width: float Width of the arms of the cross
        layer: int, array-like[2], or set Specific layer(s) to put polygon geometry on

    .. plot::
      :include-source:

      import pp

      c = pp.c.cross(length=10, width=3, layer=0)
      pp.plotgds(c)

    """

    c = pp.Component()
    R = pp.c.rectangle(size=(width, length), layer=layer)
    r1 = c.add_ref(R).rotate(90)
    r2 = c.add_ref(R)
    r1.center = (0, 0)
    r2.center = (0, 0)
    return c


if __name__ == "__main__":
    c = cross()
    pp.show(c)



# NOTE: import order matters. Only change the order if you know what you are doing
# level 0 components
from pp.components.waveguide import waveguide
from pp.components.waveguide_heater import waveguide_heater
from pp.components.waveguide_heater import wg_heater_connected

from pp.components.bend_circular import bend_circular
from pp.components.bend_circular import bend_circular180
from pp.components.bend_circular_heater import bend_circular_heater
from pp.components.bend_s import bend_s
from pp.components.bezier import bezier
from pp.components.euler.bend_euler import bend_euler90
from pp.components.euler.bend_euler import bend_euler180

from pp.components.coupler90 import coupler90
from pp.components.coupler_straight import coupler_straight
from pp.components.coupler_symmetric import coupler_symmetric
from pp.components.coupler_asymmetric import coupler_asymmetric
from pp.components.hline import hline

# basic shapes
from pp.components.circle import circle
from pp.components.compass import compass
from pp.components.cross import cross
from pp.components.crossing_waveguide import crossing
from pp.components.ellipse import ellipse
from pp.components.label import label
from pp.components.rectangle import rectangle
from pp.components.rectangle import rectangle_centered
from pp.components.ring import ring
from pp.components.taper import taper
from pp.components.taper import taper_strip_to_ridge
from pp.components.text import text

# optical test structures
from pp.components.pcm.litho_calipers import litho_calipers
from pp.components.pcm.litho_star import litho_star
from pp.components.pcm.litho_steps import litho_steps

from pp.components.grating_coupler.elliptical import grating_coupler_elliptical_te
from pp.components.grating_coupler.elliptical import grating_coupler_elliptical_tm
from pp.components.grating_coupler.uniform import grating_coupler_uniform
from pp.components.grating_coupler.grating_coupler_tree import grating_coupler_tree
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_te
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_tm
from pp.components.delay_snake import delay_snake
from pp.components.spiral_inner_io import spiral_inner_io_euler
from pp.components.spiral_inner_io import spiral_inner_io
from pp.components.spiral_external_io import spiral_external_io
from pp.components.spiral_circular import spiral_circular

# electrical
from pp.components.electrical.pad import pad
from pp.components.electrical.pad import pad_array

# electrical PCM
from pp.components.pcm.test_resistance import test_resistance
from pp.components.pcm.test_via import test_via

# level 1 components
from pp.components.coupler import coupler
from pp.components.ring_single_bus import ring_single_bus
from pp.components.ring_double_bus import ring_double_bus
from pp.components.mmi1x2 import mmi1x2
from pp.components.mmi2x2 import mmi2x2
from pp.components.mzi2x2 import mzi_arm
from pp.components.mzi2x2 import mzi2x2
from pp.components.mzi1x2 import mzi1x2
from pp.components.mzi import mzi
from pp.components.loop_mirror import loop_mirror

# level 2 components
from pp.components.component_lattice import component_lattice


# we test each factory component hash against the GDS saved in the library """
component_type2factory = {
    "bend_circular": bend_circular,
    "bend_circular180": bend_circular180,
    "bend_circular_heater": bend_circular_heater,
    "bend_euler90": bend_euler90,
    "bend_euler180": bend_euler180,
    "grating_coupler_elliptical_te": grating_coupler_elliptical_te,
    "grating_coupler_elliptical_tm": grating_coupler_elliptical_tm,
    "grating_coupler_tree": grating_coupler_tree,
    "bend_s": bend_s,
    "compass": compass,
    "coupler": coupler,
    "coupler90": coupler90,
    "coupler_straight": coupler_straight,
    "coupler_symmetric": coupler_symmetric,
    "coupler_asymmetric": coupler_asymmetric,
    "grating_coupler_uniform": grating_coupler_uniform,
    "litho_calipers": litho_calipers,
    "litho_star": litho_star,
    "litho_steps": litho_steps,
    "loop_mirror": loop_mirror,
    "mmi1x2": mmi1x2,
    "mmi2x2": mmi2x2,
    "mzi": mzi,
    "mzi2x2": mzi2x2,
    "mzi1x2": mzi1x2,
    "pad": pad,
    "pad_array": pad_array,
    "rectangle": rectangle,
    "ring": ring,
    "ring_single_bus": ring_single_bus,
    "ring_double_bus": ring_double_bus,
    "spiral_inner_io_euler": spiral_inner_io_euler,
    "spiral_inner_io": spiral_inner_io,
    "spiral_external_io": spiral_external_io,
    "taper": taper,
    "waveguide": waveguide,
    "waveguide_heater": waveguide_heater,
    "wg_heater_connected": wg_heater_connected,
}


def component_factory(component_type, **settings):
    """ returns a component with settings """
    import pp

    if isinstance(component_type, pp.Component):
        return component_type
    elif callable(component_type):
        return component_type(**settings)
    elif component_type not in component_type2factory.keys():
        raise ValueError(
            "component types available: \n {}".format(
                "\n".join(component_type2factory.keys())
            )
        )
    return component_type2factory[component_type](**settings)


# all this components are available as pp.c
__all__ = [
    "bend_circular",
    "bend_circular180",
    "bend_circular_heater",
    "bend_euler90",
    "bend_euler180",
    "bend_s",
    "bezier",
    "compass",
    "cross",
    "ring",
    "ellipse",
    "circle",
    "coupler",
    "coupler90",
    "coupler_straight",
    "coupler_symmetric",
    "coupler_asymmetric",
    "crossing",
    "delay_snake",
    "grating_coupler_uniform",
    "grating_coupler_te",
    "grating_coupler_tm",
    "grating_coupler_elliptical_te",
    "grating_coupler_elliptical_tm",
    "grating_coupler_tree",
    "hline",
    "label",
    "litho_calipers",
    "litho_star",
    "litho_steps",
    "loop_mirror",
    "mmi1x2",
    "mmi2x2",
    "mzi_arm",
    "mzi",
    "pad",
    "pad_array",
    "rectangle",
    "rectangle_centered",
    "ring_single_bus",
    "ring_double_bus",
    "spiral_circular",
    "spiral_inner_io",
    "spiral_inner_io_euler",
    "spiral_external_io",
    "taper",
    "taper_strip_to_ridge",
    "text",
    "test_via",
    "test_resistance",
    "waveguide",
    "waveguide_heater",
    "wg_heater_connected",
    "component_lattice",
]

leaf_components = ["bend_circular", "bend_euler90", "coupler", "mmi1x2", "mmi2x2"]

_skip_test = set(["label"])
_skip_test_ports = set(["grating_coupler_tree"])


def write_test_properties():
    """ writes a regression test for all the component properties dict"""
    with open("test_components.py", "w") as f:
        f.write(
            "# this code has been automatically generated from pp/components/__init__.py\n"
        )
        f.write("import pp\n\n")

        for c in set(__all__) - _skip_test:
            f.write(
                f"""
def test_{c}(data_regression):
    c = pp.c.{c}()
    data_regression.check(c.get_settings())

"""
            )


def write_test_ports():
    """ writes a regression test for all the ports """
    with open("test_ports.py", "w") as f:
        f.write(
            "# this code has been automatically generated from pp/components/__init__.py\n"
        )
        f.write("import pp\n\n")

        for component_function in (
            set(component_type2factory.values()) - _skip_test_ports
        ):
            c = component_function.__name__
            if component_function().ports:

                f.write(
                    f"""
def test_{c}(num_regression):
    c = pp.c.{c}()
    num_regression.check(c.get_ports_array())

    """
                )


if __name__ == "__main__":
    write_test_properties()
    write_test_ports()



import uuid
import pp


def connect_electrical_shortest_path(port1, port2):
    """ needs some work

    does not work for some cases

    """
    c = pp.Component(name="zz_conn_{}".format(uuid.uuid4()))
    layer = port1.layer
    p1x0 = port1.endpoints[0][0]
    p1y0 = port1.endpoints[0][1]
    p1x1 = port1.endpoints[1][0]
    p1y1 = port1.endpoints[1][1]

    p2x0 = port2.endpoints[0][0]
    p2y0 = port2.endpoints[0][1]
    p2x1 = port2.endpoints[1][0]
    p2y1 = port2.endpoints[1][1]
    c.add_polygon(
        ([(p1x0, p1y0), (p1x1, p1y1), (p2x1, p2y1), (p2x0, p2y0)]), layer=layer
    )
    return c.ref()


if __name__ == "__main__":
    from pp.components.electrical.pad import pad

    c = pp.Component()
    p = pad()
    p1 = p.ref(position=(0, 0))
    p2 = p.ref(position=(200, 0))
    c.add(p1)
    c.add(p2)
    route = connect_electrical_shortest_path(port1=p1.ports["E"], port2=p2.ports["W"])
    c.add(route)
    pp.show(c)



""" add markers to each port

"""
import numpy as np
from pp.layers import LAYER
import pp


def _rotate(v, m):
    return np.dot(m, v)


def add_pin_triangle(component, port, port_layer=LAYER.PORT, label_layer=LAYER.TEXT):
    """
    # The port visualization pattern is a triangle with a right angle
    # The face opposite the right angle is the port width
    """
    p = port

    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2

    dbot = np.array([0, -d])
    dtop = np.array([0, d])
    dtip = np.array([d, 0])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptip = p.position + _rotate(dtip, rot_mat)
    polygon = [p0, p1, ptip]

    component.add_label(
        text=p.name, position=p.midpoint, layer=label_layer,
    )

    component.add_polygon(polygon, layer=port_layer)


def add_pin_square_inside(
    component, port, port_length=0.1, port_layer=LAYER.PORT, label_layer=LAYER.TEXT
):
    """
    square inside
    """
    p = port
    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2
    dx = port_length

    dbot = np.array([0, -d])
    dtop = np.array([0, d])
    dbotin = np.array([-dx, -d])
    dtopin = np.array([-dx, +d])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=port_layer)


def add_pin_square(
    component, port, port_length=0.1, port_layer=LAYER.PORT, label_layer=LAYER.PORT
):
    """
    half out
    """
    p = port
    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2
    dx = port_length

    dbot = np.array([dx / 2, -d])
    dtop = np.array([dx / 2, d])
    dbotin = np.array([-dx / 2, -d])
    dtopin = np.array([-dx / 2, +d])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=port_layer)

    component.add_label(
        text=str(p.name), position=p.midpoint, layer=label_layer,
    )


def add_pins(
    component, add_port_marker_function=add_pin_square, add_outline=True, **kwargs,
):

    """ add port markers:

    - rectangle
    - triangle

    Add device recognition layer

    """
    if add_outline:
        c = component
        points = [
            [c.xmin, c.ymin],
            [c.xmax, c.ymin],
            [c.xmax, c.ymax],
            [c.xmin, c.ymax],
        ]
        c.add_polygon(points, layer=LAYER.DEVREC)

    if hasattr(component, "ports") and component.ports:
        for p in component.ports.values():
            add_port_marker_function(component=component, port=p, **kwargs)


if __name__ == "__main__":
    # from pp.components import mmi1x2
    # from pp.components import bend_circular
    # from pp.add_grating_couplers import add_grating_couplers

    # c = mmi1x2(width_mmi=5)
    # c = bend_circular()
    # cc = add_grating_couplers(c, layer_label=pp.LAYER.LABEL)

    # c = pp.c.waveguide()
    c = pp.c.crossing()
    # add_pins(c)
    pp.show(c)



from collections import namedtuple

layermap = dict(WG=(1, 0), DEVREC=(68, 0), LABEL=(10, 0), PORT=(1, 10))

LAYER = namedtuple("layer", layermap.keys())(*layermap.values())


if __name__ == "__main__":
    print(LAYER)



""" gdsfactory loads a configuration from 3 files, high priority overwrites low priority:

1. A config.yml found in the current working directory (highest priority)
2. ~/.gdsfactory/config.yml specific for the machine
3. the default_config is in this file (lowest priority)

"""

__version__ = "1.1.8"
__all__ = ["CONFIG", "load_config", "write_config"]

import os
import json
import subprocess
import pathlib
import ast
import logging
from pprint import pprint

import hiyapyco
import numpy as np
from git import Repo

from pp.layers import LAYER


default_config = """
tech: generic
cache_url:
BBOX_LAYER_EXCLUDE: "[]"
with_settings_label: True
add_pins: True
"""

home = pathlib.Path.home()
cwd = pathlib.Path.cwd()
roots = [pathlib.Path("/"), pathlib.Path("C:\\")]

module_path = pathlib.Path(__file__).parent.absolute()
repo_path = module_path.parent
home_path = pathlib.Path.home() / ".gdsfactory"
home_path.mkdir(exist_ok=True)

cwd_config = cwd / "config.yml"
home_config = home_path / "config.yml"


def load_config(cwd_config=cwd_config):
    """ loads config.yml and returns a dict with the config """
    cwd = cwd_config.parent
    # Find cwd config going up recursively
    while cwd not in roots:
        cwd_config = cwd / "config.yml"
        if os.path.exists(cwd_config):
            break
        cwd = cwd.parent

        if str(cwd).count("\\") <= 1 and str(cwd).endswith("\\"):
            """
            Ensure the loop terminates on a windows machine
            """
            break

    CONFIG = hiyapyco.load(
        default_config,
        str(home_config),
        str(cwd_config),
        failonmissingfiles=False,
        loglevelmissingfiles=logging.DEBUG,
    )

    CONFIG["config_path"] = cwd_config.absolute()
    CONFIG["repo_path"] = repo_path
    CONFIG["module_path"] = module_path
    CONFIG["font_path"] = module_path / "gds" / "alphabet.gds"
    CONFIG["masks_path"] = repo_path / "mask"
    CONFIG["version"] = __version__
    CONFIG["home"] = home
    CONFIG["cwd"] = cwd

    mask_name = "notDefined"

    if CONFIG.get("mask"):
        mask_name = CONFIG["mask"]["name"]
        mask_config_directory = cwd
        build_directory = mask_config_directory / "build"
        CONFIG["devices_directory"] = mask_config_directory / "devices"
        CONFIG["mask"]["gds"] = (
            mask_config_directory / "build" / "mask" / (mask_name + ".gds")
        )
    else:
        build_directory = home_path / "build"
        mask_config_directory = home_path / "build"

    if "custom_components" not in CONFIG:
        CONFIG["custom_components"] = None

    if "gdslib" not in CONFIG:
        CONFIG["gdslib"] = repo_path / "gdslib"
    CONFIG["gdslib_test"] = home_path / "gdslib_test"

    CONFIG["build_directory"] = build_directory
    CONFIG["log_directory"] = build_directory / "log"
    CONFIG["gds_directory"] = build_directory / "devices"
    CONFIG["cache_doe_directory"] = build_directory / "cache_doe"
    CONFIG["doe_directory"] = build_directory / "doe"
    CONFIG["mask_directory"] = build_directory / "mask"
    CONFIG["mask_gds"] = build_directory / "mask" / (mask_name + ".gds")
    CONFIG["mask_config_directory"] = mask_config_directory
    CONFIG["gdspath"] = build_directory / "gds.gds"
    CONFIG["samples_path"] = module_path / "samples"
    CONFIG["components_path"] = module_path / "components"

    if "gds_resources" in CONFIG:
        CONFIG["gds_resources"] = CONFIG["masks_path"] / CONFIG["gds_resources"]

    build_directory.mkdir(exist_ok=True)
    CONFIG["log_directory"].mkdir(exist_ok=True)
    CONFIG["gds_directory"].mkdir(exist_ok=True)
    CONFIG["doe_directory"].mkdir(exist_ok=True)
    CONFIG["mask_directory"].mkdir(exist_ok=True)
    CONFIG["gdslib_test"].mkdir(exist_ok=True)

    CONFIG["layers"] = CONFIG.get("layers", LAYER._asdict())

    return CONFIG


def complex_encoder(z):
    if isinstance(z, pathlib.Path):
        return str(z)
    else:
        type_name = type(z)
        raise TypeError(f"Object {z} of type {type_name} is not serializable")


def write_config(config, json_out_path):
    with open(json_out_path, "w") as f:
        json.dump(config, f, indent=2, sort_keys=True, default=complex_encoder)


CONFIG = load_config()
CONFIG["grid_unit"] = 1e-6
CONFIG["grid_resolution"] = 1e-9
CONFIG["bend_radius"] = 10.0

try:
    CONFIG["git_hash"] = Repo(repo_path).head.object.hexsha
except Exception:
    CONFIG["git_hash"] = __version__


def print_config(key=None):
    if key:
        if CONFIG.get(key):
            print(CONFIG[key])
        else:
            print(f"`{key}` key not found in {cwd_config}")
    else:
        pprint(CONFIG)


def call_if_func(f, **kwargs):
    return f(**kwargs) if callable(f) else f


def parse_layer_exclude(layer):
    return list(ast.literal_eval(layer))


def get_git_hash():
    """ Get the current git hash """
    try:
        with open(os.devnull, "w") as shutup:
            return (
                subprocess.check_output(["git", "rev-parse", "HEAD"], stderr=shutup)
                .decode("utf-8")
                .strip("\n")
            )
    except subprocess.CalledProcessError:
        return "not_a_git_repo"


GRID_UNIT = CONFIG["grid_unit"]
GRID_RESOLUTION = CONFIG["grid_resolution"]

GRID_PER_UNIT = GRID_UNIT / GRID_RESOLUTION

GRID_ROUNDING_RESOLUTION = int(np.log10(GRID_PER_UNIT))
BEND_RADIUS = CONFIG["bend_radius"]

WG_EXPANDED_WIDTH = 2.5
TAPER_LENGTH = 35.0


CONFIG["BBOX_LAYER_EXCLUDE"] = parse_layer_exclude(CONFIG["BBOX_LAYER_EXCLUDE"])

if __name__ == "__main__":
    # print_config("gdslib")
    print_config()
    # print(CONFIG["git_hash"])
    # print(CONFIG)



from numpy import floor
import pp


@pp.autoname
def via(width=0.7, height=0.7, period=2.0, clearance=1.0, layer=pp.LAYER.VIA1):
    """
    Square via
    """

    cmp = pp.Component()
    cmp.info["period"] = period
    cmp.info["clearance"] = clearance
    cmp.info["width"] = width
    cmp.info["height"] = height

    a = width / 2
    b = height / 2

    cmp.add_polygon([(-a, -b), (a, -b), (a, b), (-a, b)], layer=layer)

    return cmp


@pp.autoname
def via1(**kwargs):
    return via(layer=pp.LAYER.VIA1, **kwargs)


@pp.autoname
def via2(**kwargs):
    return via(layer=pp.LAYER.VIA2, **kwargs)


@pp.autoname
def via3(**kwargs):
    return via(layer=pp.LAYER.VIA3, **kwargs)


@pp.autoname
def tlm(
    width=11.0,
    height=11.0,
    layers=[pp.LAYER.M1, pp.LAYER.M2, pp.LAYER.M3],
    vias=[via2, via3],
):
    """
    Rectangular transition thru metal layers

    Args:
        name: component name
        width, height: rectangle parameters
        layers: layers on which to draw rectangles
        vias: vias to use to fill the rectangles

    Returns
        <pp.Component>
    """

    # assert len(layers) - 1 == len(vias), "tlm: There should be N layers for N-1 vias"

    a = width / 2
    b = height / 2
    rect_pts = [(-a, -b), (a, -b), (a, b), (-a, b)]

    cmp = pp.Component()
    # Add metal rectangles
    for layer in layers:
        cmp.add_polygon(rect_pts, layer=layer)

    # Add vias
    for via in vias:
        via = pp.call_if_func(via)

        w = via.info["width"]
        h = via.info["height"]
        c = via.info["clearance"]
        period = via.info["period"]

        nb_vias_x = (width - w - 2 * c) / period + 1
        nb_vias_y = (height - h - 2 * c) / period + 1

        nb_vias_x = int(floor(nb_vias_x))
        nb_vias_y = int(floor(nb_vias_y))

        cw = (width - (nb_vias_x - 1) * period - w) / 2
        ch = (height - (nb_vias_y - 1) * period - h) / 2

        x0 = -a + cw + w / 2
        y0 = -b + ch + h / 2

        for i in range(nb_vias_x):
            for j in range(nb_vias_y):
                cmp.add(via.ref(position=(x0 + i * period, y0 + j * period)))

    return cmp


if __name__ == "__main__":
    c = tlm()
    pp.show(c)



from pp.add_grating_couplers import add_te
from pp.add_grating_couplers import add_tm

from pp.routing.connect_component import add_io_optical_te
from pp.routing.connect_component import add_io_optical_tm

from pp.add_termination import add_gratings_and_loop_back_te
from pp.add_termination import add_gratings_and_loop_back_tm

name2function = {
    "add_grating_couplers_te": add_te,
    "add_grating_couplers_tm": add_tm,
    "add_io_optical_te": add_io_optical_te,
    "add_io_optical_tm": add_io_optical_tm,
    "add_gratings_and_loop_back_te": add_gratings_and_loop_back_te,
    "add_gratings_and_loop_back_tm": add_gratings_and_loop_back_tm,
}


if __name__ == "__main__":
    print(name2function)



"""
a container creates a new component that contains the original component inside with some extra elements:

it makes sure that some of the important settings are copied from the original component to the new component

"""

import functools
from inspect import signature
import pytest
import pp


def container(component_function):
    """ decorator for creating a new component that copies properties from the original component

    - polarization
    - wavelength
    - test_protocol
    - data_analysis_protocol

    Functions decorated with container will return a new component

    .. plot::
      :include-source:

      import pp

      @pp.container
      def add_padding(component, suffix='p', layer=pp.LAYER.M1):
          c = pp.Component(name=f"{component.name}_{suffix}")
          c << component
          w, h = component.xsize, component.ysize
          points = [[w, h], [w, 0], [0, 0], [0, h]]
          c.add_polygon(points, layer=layer)
          return c

      c = pp.c.waveguide()
      cp = add_padding(c)
      pp.plotgds(c)

    """

    @functools.wraps(component_function)
    def wrapper(*args, **kwargs):
        old = None
        if "component" not in kwargs:
            if args and isinstance(args[0], pp.Component):
                old = args[0]
            else:
                raise ValueError(
                    f"containers require a component keyword argument, or first non keyword argument"
                )
        old = old or kwargs.get("component")
        new = component_function(*args, **kwargs)

        sig = signature(component_function)
        new.settings.update(**{p.name: p.default for p in sig.parameters.values()})
        new.settings["component"] = old.settings.copy()
        new.settings["component_name"] = old.name
        new.test_protocol = new.test_protocol or old.test_protocol.copy()
        new.data_analysis_protocol = (
            new.data_analysis_protocol or old.data_analysis_protocol.copy()
        )
        new.wavelength = new.wavelength or old.wavelength
        new.polarization = new.polarization or old.polarization
        return new

    return wrapper


@container
def add_padding(
    component, padding=50, x=None, y=None, layers=[pp.LAYER.PADDING], suffix="p"
):
    """ adds padding layers to component"""
    c = pp.Component(name=f"{component.name}_{suffix}")
    c << component
    x = x or padding
    y = y or padding
    points = [
        [c.xmin - x, c.ymin - y],
        [c.xmax + x, c.ymin - y],
        [c.xmax + x, c.ymax + y],
        [c.xmin - x, c.ymax + y],
    ]
    for layer in layers:
        c.add_polygon(points, layer=layer)
    return c


def test_container():
    old = pp.c.waveguide()
    suffix = "p"
    name = f"{old.name}_{suffix}"
    new = add_padding(component=old, suffix=suffix)
    assert new != old, f"new component {new} should be different from {old}"
    assert new.name == name, f"new name {new.name} should be {name}"
    # assert len(new.ports) == len(
    #     old.ports
    # ), f"new component {len(new.ports)} ports should match original {len(old.ports)} ports"
    # assert len(new.settings) == len(
    #     old.settings
    # ), f"new component {new.settings} settings should match original {old.settings} settings"
    return new


def test_container2():
    old = pp.c.waveguide()
    suffix = "p"
    name = f"{old.name}_{suffix}"
    new = add_padding(old, suffix=suffix)
    assert new != old, f"new component {new} should be different from {old}"
    assert new.name == name, f"new name {new.name} should be {name}"
    # assert len(new.ports) == len(
    #     old.ports
    # ), f"new component {len(new.ports)} ports should match original {len(old.ports)} ports"
    return new


def test_container_error():
    old = pp.c.waveguide()
    with pytest.raises(ValueError):
        add_padding(component2=old)  # will raise an error
    return old


if __name__ == "__main__":
    c = test_container2()
    pp.show(c)



import os
import gdspy

from phidl.device_layout import DeviceReference

import pp
from pp.component import NAME_TO_DEVICE


def import_gds(
    filename, cellname=None, flatten=False, overwrite_cache=False, snap_to_grid_nm=False
):
    """ returns a Componenent from a GDS file
    """
    filename = str(filename)
    gdsii_lib = gdspy.GdsLibrary()
    gdsii_lib.read_gds(filename)
    top_level_cells = gdsii_lib.top_level()

    if cellname is not None:
        if cellname not in gdsii_lib.cell_dict:
            raise ValueError(
                "[PHIDL] import_gds() The requested cell (named %s) is not present in file %s"
                % (cellname, filename)
            )
        topcell = gdsii_lib.cell_dict[cellname]
    elif cellname is None and len(top_level_cells) == 1:
        topcell = top_level_cells[0]
    elif cellname is None and len(top_level_cells) > 1:
        raise ValueError(
            "[PHIDL] import_gds() There are multiple top-level cells in {}, you must specify `cellname` to select of one of them among {}".format(
                filename, [_c.name for _c in top_level_cells]
            )
        )

    if flatten:
        D = pp.Component()
        polygons = topcell.get_polygons(by_spec=True)

        for layer_in_gds, polys in polygons.items():
            D.add_polygon(polys, layer=layer_in_gds)
        return D

    else:
        D_list = []
        c2dmap = {}
        all_cells = topcell.get_dependencies(True)
        all_cells.update([topcell])

        for cell in all_cells:
            cell_name = cell.name
            if overwrite_cache or cell_name not in NAME_TO_DEVICE:
                D = pp.Component()
                D.name = cell.name
                D.polygons = cell.polygons
                D.references = cell.references
                D.name = cell_name
                D.labels = cell.labels
            else:
                D = NAME_TO_DEVICE[cell_name]

            c2dmap.update({cell_name: D})
            D_list += [D]

        for D in D_list:
            # First convert each reference so it points to the right Device
            converted_references = []
            for e in D.references:
                try:
                    ref_device = c2dmap[e.ref_cell.name]

                    dr = DeviceReference(
                        device=ref_device,
                        origin=e.origin,
                        rotation=e.rotation,
                        magnification=e.magnification,
                        x_reflection=e.x_reflection,
                    )
                    converted_references.append(dr)
                except Exception:
                    print("WARNING - Could not import", e.ref_cell.name)

            D.references = converted_references
            # Next convert each Polygon
            temp_polygons = list(D.polygons)
            D.polygons = []
            for p in temp_polygons:
                if snap_to_grid_nm:
                    points_on_grid = pp.drc.snap_to_grid(
                        p.polygons[0], nm=snap_to_grid_nm
                    )
                    p = gdspy.Polygon(
                        points_on_grid, layer=p.layers[0], datatype=p.datatypes[0]
                    )
                D.add_polygon(p)
                # else:
                #     warnings.warn('[PHIDL] import_gds(). Warning an element which was not a ' \
                #         'polygon or reference exists in the GDS, and was not able to be imported. ' \
                #         'The element was a: "%s"' % e)

        topdevice = c2dmap[topcell.name]
        return topdevice


if __name__ == "__main__":

    filename = os.path.join(pp.CONFIG["gdslib"], "mzi2x2.gds")
    filename = pp.CONFIG["gdslib"] / "mzi2x2.gds"
    c = import_gds(filename, snap_to_grid_nm=5)
    print(c)
    pp.show(c)



""" read port markers from a GDS

"""

import csv
from pp import klive
import phidl.geometry as pg


def read_port_markers(gdspath, layer=69):
    """ loads a GDS and read port

    Args:
        gdspath:
        layer: GDS layer
    """
    D = pg.import_gds(gdspath)
    D = pg.extract(D, layers=[layer])
    for e in D.elements:
        print(e.x, e.y)


def csv2port(csvpath):
    """ loads and reads ports from a CSV file
    returns a dict
    """
    ports = {}
    with open(csvpath, "r") as csvfile:
        rows = csv.reader(csvfile, delimiter=",", quotechar="|")
        for row in rows:
            ports[row[0]] = row[1:]

    return ports


if __name__ == "__main__":
    import os
    from pp import CONFIG

    name = "mmi1x2_WM1"
    gdspath = os.path.join(CONFIG["lib"], name, name + ".gds")
    csvpath = os.path.join(CONFIG["lib"], name, name + ".ports")
    klive.show(gdspath)
    # read_port_markers(gdspath, layer=66)
    p = csv2port(csvpath)
    print(p)



import pp
from pp.add_labels import get_optical_text
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_te
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_tm
from pp.config import CONFIG


def add_grating_couplers(
    component,
    grating_coupler=grating_coupler_te,
    layer_label=CONFIG["layers"]["LABEL"],
    input_port_indexes=[0],
):
    """ returns component with grating ports and labels on each port """
    component = pp.call_if_func(component)
    c = pp.Component(name=component.name + "_c")
    c.add_ref(component)
    grating_coupler = pp.call_if_func(grating_coupler)

    for i, port in enumerate(component.ports.values()):
        t_ref = c.add_ref(grating_coupler)
        t_ref.connect(list(t_ref.ports.values())[0], port)
        label = get_optical_text(port, grating_coupler, i)
        c.label(label, position=port.midpoint, layer=layer_label)

    return c


def add_te(*args, **kwargs):
    return add_grating_couplers(*args, **kwargs)


def add_tm(*args, grating_coupler=grating_coupler_tm, **kwargs):
    return add_grating_couplers(*args, grating_coupler=grating_coupler, **kwargs)


if __name__ == "__main__":
    # c = pp.c.grating_coupler_elliptical_te()
    # print(c.wavelength)

    # print(c.get_property('wavelength'))

    c = pp.c.waveguide(width=2)
    # cc = add_grating_couplers(c)
    cc = add_tm(c)
    print(cc)
    pp.show(cc)



import uuid
from copy import deepcopy
import numpy as np
from numpy import pi, sin, cos, mod

import gdspy

from phidl.device_layout import Label
from phidl.device_layout import Device
from phidl.device_layout import DeviceReference
from phidl.device_layout import _parse_layer
from phidl.device_layout import Port as PortPhidl

from pp.ports import select_optical_ports
from pp.ports import select_electrical_ports
import pp
from pp.config import CONFIG
from pp.compare_cells import hash_cells

NAME_TO_DEVICE = {}

BBOX_LAYER_EXCLUDE = CONFIG["BBOX_LAYER_EXCLUDE"]


class Port(PortPhidl):
    """ Extends phidl port by adding layer and a port_type (optical, electrical)

    Args:
        name: we name ports according to orientation (S0, S1, W0, W1, N0 ...)
        midpoint: (0, 0)
        width: of the port
        orientation: 0
        parent: None, parent component (component to which this port belong to)
        layer: 1
        port_type: optical, dc, rf, detector, superconducting, trench

    """

    _next_uid = 0

    def __init__(
        self,
        name=None,
        midpoint=(0, 0),
        width=1,
        orientation=0,
        parent=None,
        layer=1,
        port_type="optical",
    ):
        self.name = name
        self.midpoint = np.array(midpoint, dtype="float64")
        self.width = width
        self.orientation = mod(orientation, 360)
        self.parent = parent
        self.info = {}
        self.uid = Port._next_uid
        self.layer = layer
        self.port_type = port_type

        if self.width < 0:
            raise ValueError("[PHIDL] Port creation error: width must be >=0")
        self._next_uid += 1

    def __repr__(self):
        return "Port (name {}, midpoint {}, width {}, orientation {}, layer {}, port_type {})".format(
            self.name,
            self.midpoint,
            self.width,
            self.orientation,
            self.layer,
            self.port_type,
        )

    @property
    def angle(self):
        """convenient alias"""
        return self.orientation

    @angle.setter
    def angle(self, a):
        self.orientation = a

    @property
    def position(self):
        return self.midpoint

    @position.setter
    def position(self, p):
        self.midpoint = np.array(p, dtype="float64")

    def move(self, vector):
        self.midpoint = self.midpoint + np.array(vector)

    def move_polar_copy(self, d, angle):
        port = self._copy()
        DEG2RAD = np.pi / 180
        dp = np.array((d * np.cos(DEG2RAD * angle), d * np.sin(DEG2RAD * angle)))
        self.move(dp)
        return port

    def flip(self):
        """ flips port """
        port = self._copy()
        port.angle = (port.angle + 180) % 360
        return port

    def _copy(self, new_uid=True):
        new_port = Port(
            name=self.name,
            midpoint=self.midpoint,
            width=self.width,
            orientation=self.orientation,
            parent=self.parent,
            layer=self.layer,
            port_type=self.port_type,
        )
        new_port.info = deepcopy(self.info)
        if not new_uid:
            new_port.uid = self.uid
            Port._next_uid -= 1
        return new_port


class SizeInfo:
    def __init__(self, bbox):
        self.west = bbox[0, 0]
        self.east = bbox[1, 0]
        self.south = bbox[0, 1]
        self.north = bbox[1, 1]

        self.width = self.east - self.west
        self.height = self.north - self.south

        xc = 0.5 * (self.east + self.west)
        yc = 0.5 * (self.north + self.south)

        self.sw = np.array([self.west, self.south])
        self.se = np.array([self.east, self.south])
        self.nw = np.array([self.west, self.north])
        self.ne = np.array([self.east, self.north])

        self.cw = np.array([self.west, yc])
        self.ce = np.array([self.east, yc])
        self.nc = np.array([xc, self.north])
        self.sc = np.array([xc, self.south])
        self.cc = self.center = np.array([xc, yc])

    def get_rect(
        self, padding=0, padding_w=None, padding_e=None, padding_n=None, padding_s=None
    ):
        w, e, s, n = self.west, self.east, self.south, self.north

        padding_n = padding if padding_n is None else padding_n
        padding_e = padding if padding_e is None else padding_e
        padding_w = padding if padding_w is None else padding_w
        padding_s = padding if padding_s is None else padding_s

        w = w - padding_w
        e = e + padding_e
        s = s - padding_s
        n = n + padding_n

        return [(w, s), (e, s), (e, n), (w, n)]

    @property
    def rect(self):
        return self.get_rect()

    def __str__(self):
        return "w: {}\ne: {}\ns: {}\nn: {}\n".format(
            self.west, self.east, self.south, self.north
        )


def _rotate_points(points, angle=45, center=(0, 0)):
    """ Rotates points around a centerpoint defined by ``center``.  ``points`` may be
    input as either single points [1,2] or array-like[N][2], and will return in kind
    """
    # First check for common, easy values of angle
    p_arr = np.asarray(points)
    if angle == 0:
        return p_arr

    c0 = np.asarray(center)
    displacement = p_arr - c0
    if angle == 180:
        return c0 - displacement

    if p_arr.ndim == 2:
        perpendicular = displacement[:, ::-1]
    elif p_arr.ndim == 1:
        perpendicular = displacement[::-1]

    # Fall back to trigonometry
    angle = angle * pi / 180
    ca = cos(angle)
    sa = sin(angle)
    sa = np.array((-sa, sa))
    return displacement * ca + perpendicular * sa + c0


class ComponentReference(DeviceReference):
    def __init__(
        self,
        device,
        origin=(0, 0),
        rotation=0,
        magnification=None,
        x_reflection=False,
        visual_label="",
    ):
        super().__init__(
            device=device,
            origin=origin,
            rotation=rotation,
            magnification=magnification,
            x_reflection=x_reflection,
        )
        self.parent = device
        # The ports of a DeviceReference have their own unique id (uid),
        # since two DeviceReferences of the same parent Device can be
        # in different locations and thus do not represent the same port
        self._local_ports = {
            name: port._copy(new_uid=True) for name, port in device.ports.items()
        }
        self.visual_label = visual_label

    def __repr__(self):
        return (
            'DeviceReference (parent Device "%s", ports %s, origin %s, rotation %s, x_reflection %s)'
            % (
                self.parent.name,
                list(self.ports.keys()),
                self.origin,
                self.rotation,
                self.x_reflection,
            )
        )

    def __str__(self):
        return self.__repr__()

    def __getitem__(self, val):
        """ This allows you to access an alias from the reference's parent, and receive
        a copy of the reference which is correctly rotated and translated"""
        try:
            alias_device = self.parent[val]
        except Exception:
            raise ValueError(
                '[PHIDL] Tried to access alias "%s" from parent '
                'Device "%s", which does not exist' % (val, self.parent.name)
            )
        new_reference = ComponentReference(
            alias_device.parent,
            origin=alias_device.origin,
            rotation=alias_device.rotation,
            magnification=alias_device.magnification,
            x_reflection=alias_device.x_reflection,
        )

        if self.x_reflection:
            new_reference.reflect((1, 0))
        if self.rotation is not None:
            new_reference.rotate(self.rotation)
        if self.origin is not None:
            new_reference.move(self.origin)

        return new_reference

    def get_labels(self, recursive=True, associate_visual_labels=True):
        """
        access all labels correctly rotated, mirrored and translated
        """
        # params = {
        #     "recursive": recursive,
        #     "associate_visual_labels": associate_visual_labels,
        # }

        labels_untransformed = self.parent.get_labels()

        labels = []
        visual_label = self.visual_label
        for lbl in labels_untransformed:
            p = lbl.position
            position, _ = self._transform_port(
                p, 0, self.origin, self.rotation, self.x_reflection
            )

            text = lbl.text
            if visual_label and associate_visual_labels:
                text += " / " + visual_label

            new_lbl = Label(
                text=text,
                position=position,
                anchor="o",
                layer=lbl.layer,
                texttype=lbl.texttype,
            )

            labels += [new_lbl]
        if labels:
            print("R", self.parent.name, len(labels))
        return labels

    @property
    def ports(self):
        """ This property allows you to access myref.ports, and receive a copy
        of the ports dict which is correctly rotated and translated"""
        for name, port in self.parent.ports.items():
            port = self.parent.ports[name]
            new_midpoint, new_orientation = self._transform_port(
                port.midpoint,
                port.orientation,
                self.origin,
                self.rotation,
                self.x_reflection,
            )
            if name not in self._local_ports:
                self._local_ports[name] = port._copy(new_uid=True)
            self._local_ports[name].midpoint = new_midpoint
            self._local_ports[name].orientation = mod(new_orientation, 360)
            self._local_ports[name].parent = self
        # Remove any ports that no longer exist in the reference's parent
        parent_names = self.parent.ports.keys()
        local_names = list(self._local_ports.keys())
        for name in local_names:
            if name not in parent_names:
                self._local_ports.pop(name)
        return self._local_ports

    @property
    def info(self):
        return self.parent.info

    # @property
    # def bbox(self):
    # bbox = self.get_bounding_box()
    # if bbox is None:
    # bbox = ((0, 0), (0, 0))
    # return np.array(bbox)

    # def get_bounding_box(self):
    # """
    # WARNING ! This does not work as well as GDSPY method. But is cheaper.
    # Comment this method if packing breaks

    # the bounding box of a reference is the transformed
    # bounding box of the parent cell
    # """

    # _bbox = self.parent.get_bounding_box()
    # bbox = np.array([self._transform_point(
    # p, self.origin, self.rotation, self.x_reflection
    # ) for p in _bbox])
    # return bbox

    def get_bounding_box2(self):
        """
        Returns the bounding box for this reference.

        Returns
        -------
        out : Numpy array[2,2] or ``None``
            Bounding box of this cell [[x_min, y_min], [x_max, y_max]],
            or ``None`` if the cell is empty.
        """
        if not isinstance(self.ref_cell, gdspy.Cell):
            return None
        if (
            self.rotation is None
            and self.magnification is None
            and self.x_reflection is None
        ):
            key = self.ref_cell
        else:
            key = (self.ref_cell, self.rotation, self.magnification, self.x_reflection)

        if key not in gdspy._bounding_boxes:
            """
            The bounding boxes are all valid, but this specific transformation
            has not been registered yet. Fetch bbox, and apply transform
            """
            _bb = self.ref_cell.get_bounding_box()
            if _bb is not None:
                bb = np.array(
                    [
                        self._transform_point(
                            p, (0, 0), self.rotation, self.x_reflection
                        )
                        for p in _bb
                    ]
                )
                _x1, _x2 = bb[0, 0], bb[1, 0]
                _y1, _y2 = bb[0, 1], bb[1, 1]

                bb[0, 0] = min(_x1, _x2)
                bb[1, 0] = max(_x1, _x2)
                bb[0, 1] = min(_y1, _y2)
                bb[1, 1] = max(_y1, _y2)
                gdspy._bounding_boxes[key] = bb

            else:
                bb = None
        else:
            bb = gdspy._bounding_boxes[key]
        if self.origin is None or bb is None:
            return bb
        else:
            translation = np.array(
                [(self.origin[0], self.origin[1]), (self.origin[0], self.origin[1])]
            )
            return bb + translation

    @property
    def bbox(self):
        _bbox = self.ref_cell.get_bounding_box()
        bbox = np.array(
            [
                self._transform_point(p, self.origin, self.rotation, self.x_reflection)
                for p in _bbox
            ]
        )

        # Rearrange min,max
        bbox[1][0], bbox[0][0] = (
            max(bbox[1][0], bbox[0][0]),
            min(bbox[1][0], bbox[0][0]),
        )
        bbox[1][1], bbox[0][1] = (
            max(bbox[1][1], bbox[0][1]),
            min(bbox[1][1], bbox[0][1]),
        )

        return bbox

    @property
    def size_info(self):
        return SizeInfo(self.bbox)
        # if self.__size_info__ == None:
        # return self.__size_info__

    def _transform_port(
        self, point, orientation, origin=(0, 0), rotation=None, x_reflection=False
    ):
        # Apply GDS-type transformations to a port (x_ref)
        new_point = np.array(point)
        new_orientation = orientation

        if x_reflection:
            new_point[1] = -new_point[1]
            new_orientation = -orientation
        if rotation is not None:
            new_point = _rotate_points(new_point, angle=rotation, center=[0, 0])
            new_orientation += rotation
        if origin is not None:
            new_point = new_point + np.array(origin)
        new_orientation = mod(new_orientation, 360)

        return new_point, new_orientation

    def _transform_point(self, point, origin=(0, 0), rotation=None, x_reflection=False):
        # Apply GDS-type transformations to a port (x_ref)
        new_point = np.array(point)

        if x_reflection:
            new_point[1] = -new_point[1]
        if rotation is not None:
            new_point = _rotate_points(new_point, angle=rotation, center=[0, 0])
        if origin is not None:
            new_point = new_point + np.array(origin)

        return new_point

    def move(self, origin=(0, 0), destination=None, axis=None):
        """ Moves the DeviceReference from the origin point to the destination.  Both
         origin and destination can be 1x2 array-like, Port, or a key
         corresponding to one of the Ports in this device_ref """

        # If only one set of coordinates is defined, make sure it's used to move things
        if destination is None:
            destination = origin
            origin = (0, 0)

        if isinstance(origin, Port):
            o = origin.midpoint
        elif np.array(origin).size == 2:
            o = origin
        elif origin in self.ports:
            o = self.ports[origin].midpoint
        else:
            raise ValueError(
                "[ComponentReference.move()] ``origin`` not array-like, a port, or port name"
            )

        if isinstance(destination, Port):
            d = destination.midpoint
        elif np.array(destination).size == 2:
            d = destination
        elif destination in self.ports:
            d = self.ports[destination].midpoint
        else:
            raise ValueError(
                "[ComponentReference.move()] ``destination`` not array-like, a port, or port name"
            )

        # Lock one axis if necessary
        if axis == "x":
            d = (d[0], o[1])
        if axis == "y":
            d = (o[0], d[1])

        # This needs to be done in two steps otherwise floating point errors can accrue
        dxdy = np.array(d) - np.array(o)
        self.origin = np.array(self.origin) + dxdy
        self._bb_valid = False
        return self

    def rotate(self, angle=45, center=(0, 0)):
        if angle == 0:
            return self
        if type(center) == str or type(center) == int:
            center = self.ports[center].position

        if type(center) is Port:
            center = center.midpoint
        self.rotation += angle
        self.rotation = self.rotation % 360
        self.origin = _rotate_points(self.origin, angle, center)
        self._bb_valid = False
        return self

    def reflect_h(self, port_name=None, x0=None):
        """
        Perform horizontal mirror (w.r.t vertical axis)
        """
        if port_name is None and x0 is None:
            x0 = 0

        if port_name is not None:
            position = self.ports[port_name]
            x0 = position.x
        self.reflect((x0, 1), (x0, 0))

    def reflect_v(self, port_name=None, y0=None):
        """
        Perform vertical mirror (w.r.t horizontal axis)
        """
        if port_name is None and y0 is None:
            y0 = 0

        if port_name is not None:
            position = self.ports[port_name]
            y0 = position.y
        self.reflect((1, y0), (0, y0))

    def reflect(self, p1=(0, 1), p2=(0, 0)):
        if type(p1) is Port:
            p1 = p1.midpoint
        if type(p2) is Port:
            p2 = p2.midpoint
        p1 = np.array(p1)
        p2 = np.array(p2)
        # Translate so reflection axis passes through origin
        self.origin = self.origin - p1

        # Rotate so reflection axis aligns with x-axis
        angle = np.arctan2((p2[1] - p1[1]), (p2[0] - p1[0])) * 180 / pi
        self.origin = _rotate_points(self.origin, angle=-angle, center=[0, 0])
        self.rotation -= angle

        # Reflect across x-axis
        self.x_reflection = not self.x_reflection
        self.origin[1] = -self.origin[1]
        self.rotation = -self.rotation

        # Un-rotate and un-translate
        self.origin = _rotate_points(self.origin, angle=angle, center=[0, 0])
        self.rotation += angle
        self.rotation = self.rotation % 360
        self.origin = self.origin + p1

        self._bb_valid = False
        return self

    def connect(self, port, destination, overlap=0):
        # ``port`` can either be a string with the name or an actual Port
        if port in self.ports:  # Then ``port`` is a key for the ports dict
            p = self.ports[port]
        elif type(port) is Port:
            p = port
        else:
            raise ValueError(
                f"{self}.connect({port}) valid ports are {self.ports.keys()}"
            )

        angle = 180 + destination.orientation - p.orientation
        angle = angle % 360
        self.rotate(angle=angle, center=p.midpoint)

        self.move(origin=p, destination=destination)
        self.move(
            -overlap
            * np.array(
                [
                    cos(destination.orientation * pi / 180),
                    sin(destination.orientation * pi / 180),
                ]
            )
        )
        return self

    def get_property(self, property):
        if hasattr(self, property):
            return self.property

        return self.ref_cell.get_property(property)


class Component(Device):
    """ inherits from phidl.Device and extends some functionality

    - get/write JSON metadata
    - get ports by type (optical, electrical ...)
    - set data_analysis and test_protocols

    """

    def __init__(
        self,
        name="Unnamed",
        polarization=None,
        wavelength=None,
        test_protocol=None,
        data_analysis_protocol=None,
        *args,
        **kwargs,
    ):
        # Allow name to be set like Component('arc') or Component(name = 'arc')

        self.data_analysis_protocol = data_analysis_protocol or {}
        self.test_protocol = test_protocol or {}
        self.wavelength = wavelength
        self.polarization = polarization

        self.settings = kwargs
        self.__ports__ = {}
        self.info = {}
        self.aliases = {}
        self.uid = str(uuid.uuid4())[:8]

        if "with_uuid" in kwargs or name == "Unnamed":
            name += "_" + self.uid

        super(Component, self).__init__(name=name, exclude_from_current=True)
        self.name = name

    def get_optical_ports(self):
        """ returns a lit of optical ports """
        return list(select_optical_ports(self.ports).values())

    def get_ports_array(self):
        """ returns ports as a dict of np arrays"""
        return {
            port_name: np.array(
                [
                    port.x,
                    port.y,
                    int(port.orientation),
                    port.width,
                    port.layer[0],
                    port.layer[1],
                ]
            )
            for port_name, port in self.ports.items()
        }

    def get_electrical_ports(self):
        """ returns a list of optical ports """
        return list(select_electrical_ports(self.ports).values())

    def get_properties(self):
        """ returns name, uid, ports, aliases and numer of references """
        return f"name: {self._internal_name}, uid: {self.uid},  ports: {self.ports.keys()}, aliases {self.aliases.keys()}, number of references: {len(self.references)}"

    def ref(
        self, position=(0, 0), port_id=None, rotation=0, h_mirror=False, v_mirror=False
    ):
        """ returns a reference of the component """
        _ref = ComponentReference(self)

        if port_id and port_id not in self.ports:
            raise ValueError(f"port {port_id} not in {self.ports.keys()}")

        if port_id:
            origin = self.ports[port_id].position
        else:
            origin = (0, 0)

        if h_mirror:
            _ref.reflect_h(port_id)

        if v_mirror:
            _ref.reflect_v(port_id)

        if rotation != 0:
            _ref.rotate(rotation, origin)
        _ref.move(origin, position)
        return _ref

    def ref_center(self, position=(0, 0)):
        """ returns a reference of the component centered at (x=0, y=0)"""
        si = self.size_info
        yc = si.south + si.height / 2
        xc = si.west + si.width / 2
        center = (xc, yc)
        _ref = ComponentReference(self)
        _ref.move(center, position)
        return _ref

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.__repr__()

    def update_settings(self, **kwargs):
        """ update settings dict """
        for key, value in kwargs.items():
            self.settings[key] = _clean_value(value)

    def get_property(self, property):
        if property in self.settings:
            return self.settings[property]
        if hasattr(self, property):
            return getattr(self, property)

        raise ValueError(
            "Component {} does not have property {}".format(self.name, property)
        )

    def get_settings(self):
        """ Returns settings dictionary
        """
        output = {}
        ignore = set(
            dir(Component())
            + ["path", "settings", "properties", "function_name", "type", "netlist"]
        )
        params = set(dir(self)) - ignore
        output["name"] = self.name

        if hasattr(self, "function_name"):
            output["class"] = self.function_name

        for key, value in self.settings.items():
            output[key] = _clean_value(value)

        for param in params:
            value = getattr(self, param)
            output[param] = _clean_value(value)

        # output["hash"] = hashlib.md5(json.dumps(output).encode()).hexdigest()
        # output["hash_geometry"] = str(self.hash_geometry())
        return output

    def add_port(
        self,
        name=None,
        midpoint=(0, 0),
        width=1,
        orientation=45,
        port=None,
        layer=(1, 0),
        port_type="optical",
    ):
        """ Can be called to copy an existing port like add_port(port = existing_port) or
        to create a new port add_port(myname, mymidpoint, mywidth, myorientation).
        Can also be called to copy an existing port with a new name like add_port(port = existing_port, name = new_name)"""
        if port:
            if not isinstance(port, Port):
                print(type(port))
                raise ValueError(
                    "[PHIDL] add_port() error: Argument `port` must be a Port for copying"
                )
            p = port._copy(new_uid=True)
            if name is not None:
                p.name = name
            p.parent = self

        elif isinstance(name, Port):
            p = name._copy(new_uid=True)
            p.parent = self
            name = p.name
        else:
            p = Port(
                name=name,
                midpoint=midpoint,
                width=width,
                orientation=orientation,
                parent=self,
                layer=layer,
                port_type=port_type,
            )
        if name is not None:
            p.name = name
        if p.name in self.ports:
            raise ValueError(
                '[DEVICE] add_port() error: Port name "%s" already exists in this Device (name "%s", uid %s)'
                % (p.name, self._internal_name, self.uid)
            )

        self.ports[p.name] = p
        return p

    def get_json(self, **kwargs):
        """ returns JSON metadata """
        jsondata = {
            "json_version": 6,
            "cells": recurse_structures(self),
            "test_protocol": self.test_protocol,
            "data_analysis_protocol": self.data_analysis_protocol,
            "git_hash": pp.CONFIG["git_hash"],
        }

        if hasattr(self, "analysis"):
            jsondata["analysis"] = self.analysis

        return jsondata

    def hash_geometry(self):
        """ returns geometrical hash
        """
        if self.references or self.polygons:
            h = hash_cells(self, {})[self.name]
        else:
            h = "empty_geometry"

        self.settings.update(hash=h)
        return h

    def remove_layers(
        self, layers=(), include_labels=True, invert_selection=False, recursive=True
    ):
        """ remove a list of layers """
        layers = [_parse_layer(layer) for layer in layers]
        all_D = list(self.get_dependencies(recursive))
        all_D += [self]
        for D in all_D:
            for polygonset in D.polygons:
                polygon_layers = zip(polygonset.layers, polygonset.datatypes)
                polygons_to_keep = [(pl in layers) for pl in polygon_layers]
                if not invert_selection:
                    polygons_to_keep = [(not p) for p in polygons_to_keep]
                polygonset.polygons = [
                    p for p, keep in zip(polygonset.polygons, polygons_to_keep) if keep
                ]
                polygonset.layers = [
                    p for p, keep in zip(polygonset.layers, polygons_to_keep) if keep
                ]
                polygonset.datatypes = [
                    p for p, keep in zip(polygonset.datatypes, polygons_to_keep) if keep
                ]

            if include_labels:
                new_labels = []
                for label in D.labels:
                    original_layer = (label.layer, label.texttype)
                    original_layer = _parse_layer(original_layer)
                    if invert_selection:
                        keep_layer = original_layer in layers
                    else:
                        keep_layer = original_layer not in layers
                    if keep_layer:
                        new_labels += [label]
                D.labels = new_labels
        return self

    @property
    def size_info(self):
        """ size info of the component """
        # if self.__size_info__ == None:
        # self.__size_info__  = SizeInfo(self.bbox)
        return SizeInfo(self.bbox)  # self.__size_info__

    def add_ref(self, D, alias=None):
        """ Takes a Device and adds it as a ComponentReference to the current
        Device.  """
        if type(D) in (list, tuple):
            return [self.add_ref(E) for E in D]
        if not isinstance(D, Component) and not isinstance(D, Device):
            raise TypeError(
                """[PP] add_ref() was passed a {}. This is not a Component object. """.format(
                    type(D)
                )
            )
        d = ComponentReference(D)  # Create a ComponentReference (CellReference)
        self.add(d)  # Add ComponentReference (CellReference) to Device (Cell)

        if alias is not None:
            self.aliases[alias] = d
        return d

    def get_bounding_box2(self, layers_excl=BBOX_LAYER_EXCLUDE, force=False):
        """
        Calculate the bounding box for this cell.

        Returns
            out : Numpy array[2, 2] or None
                Bounding box of this cell [[x_min, y_min], [x_max, y_max]],
                or None if the cell is empty.
        """
        if (
            len(self.polygons) == 0
            and len(self.paths) == 0
            and len(self.references) == 0
        ):
            return None

        if force or (
            not (
                self._bb_valid
                and all(ref._bb_valid for ref in self.get_dependencies(True))
            )
        ):
            bb = None  # np.array(((1e300, 1e300), (-1e300, -1e300)))
            all_polygons = []
            for polygon in self.polygons:

                polygons = _filter_polys(polygon, layers_excl)

                all_polygons.extend(polygons)
            for path in self.paths:
                all_polygons.extend(path.to_polygonset().polygons)
            for reference in self.references:
                reference_bb = reference.get_bounding_box()
                if reference_bb is not None:
                    if bb is None:
                        bb = reference_bb.copy()
                    else:
                        bb[0, 0] = min(bb[0, 0], reference_bb[0, 0])
                        bb[0, 1] = min(bb[0, 1], reference_bb[0, 1])
                        bb[1, 0] = max(bb[1, 0], reference_bb[1, 0])
                        bb[1, 1] = max(bb[1, 1], reference_bb[1, 1])

            if len(all_polygons) > 0:
                all_points = np.concatenate(all_polygons).transpose()
                if bb is None:
                    bb = np.zeros([2, 2])
                    bb[0, 0] = all_points[0].min()
                    bb[0, 1] = all_points[1].min()
                    bb[1, 0] = all_points[0].max()
                    bb[1, 1] = all_points[1].max()
                else:
                    bb[0, 0] = min(bb[0, 0], all_points[0].min())
                    bb[0, 1] = min(bb[0, 1], all_points[1].min())
                    bb[1, 0] = max(bb[1, 0], all_points[0].max())
                    bb[1, 1] = max(bb[1, 1], all_points[1].max())
            self._bb_valid = True
            gdspy._bounding_boxes[self] = bb
        return gdspy._bounding_boxes[self]


def _filter_polys(polygons, layers_excl):
    return [
        p
        for p, l, d in zip(polygons.polygons, polygons.layers, polygons.datatypes)
        if (l, d) not in layers_excl
    ]


IGNORE_FUNCTION_NAMES = set()
IGNORE_STRUCTURE_NAME_PREFIXES = set(["zz_conn"])


def recurse_structures(structure):
    """ Recurse over structures """
    if (
        hasattr(structure, "function_name")
        and structure.function_name in IGNORE_FUNCTION_NAMES
    ):
        return {}

    if hasattr(structure, "name") and any(
        [structure.name.startswith(i) for i in IGNORE_STRUCTURE_NAME_PREFIXES]
    ):
        return {}
    if not hasattr(structure, "get_json"):
        return {}

    # if hasattr(structure, "function_name"):
    #     print(structure.function_name)
    output = {structure.name: structure.get_settings()}
    for element in structure.references:
        if (
            isinstance(element, ComponentReference)
            and element.ref_cell.name not in output
        ):
            output.update(recurse_structures(element.ref_cell))

    return output


def _clean_value(value):
    """ returns a clean value """
    if type(value) in [int, float, str, tuple]:
        value = value
    elif callable(value):
        value = value.__name__
    elif type(value) == pp.Component:
        value = value.name
    elif hasattr(value, "__iter__"):
        value = "_".join(["{}".format(i) for i in value]).replace(".", "p")
    # elif hasattr(value, "__iter__"):
    #     try:
    #         json.dumps(value)
    #         value = value
    #     except Exception:
    #         value = str(value)
    elif isinstance(value, np.int32):
        value = int(value)
    elif isinstance(value, np.int64):
        value = int(value)
    else:
        value = str(value)

    return value


def demo_component(port):
    c = pp.Component()
    c.add_port(name="p1", port=port)
    return c


if __name__ == "__main__":
    # c = pp.c.bend_circular180()
    # c = pp.c.coupler()
    # c.add_labels()
    # pp.show(c)

    c = pp.c.waveguide()
    # c = pp.c.mzi1x2()
    # c = pp.c.ring_double_bus()
    # print(c.hash_geometry())
    # print(c.get_json())
    # print(c.get_settings())
    # print(c.settings)

    print(c.get_settings())

    # print(json.dumps(c.get_settings()))
    # print(c.get_json()['cells'].keys())
    # print(c.get_json())

    # from pp.routing import add_io_optical
    # cc = add_io_optical(c)
    # pp.write_component(cc)

    # from pprint import pprint

    # c = pp.c.mmi1x2()
    # cc = add_io_optical(c)
    # cc.get_json()
    # pp.show(cc)
    # c.update_settings(
    #     analysis={
    #         "device_type": "loopback",
    #         "test_method": "spectrum",
    #         "analysis_keys": "[cell_name, category, 0]",
    #         "polarization_mix": "None",
    #         "device_group_type": "deembed, mapping",
    #         "optical_io_measure": "[[1,4],[2,3]]",
    #         "polarization_input": "TE",
    #         "deembed_device_group_id": "None",
    #     }
    # )

    # w = pp.c.waveguide()
    # c = demo_component(port=w.ports["E0"])
    # pprint(c.get_json())
    # pprint(c.get_settings())

    # print(c.get_json())
    # print(c.get_settings())
    # print(c.get_settings(test="hi"))
    # pp.show(c)



import numpy as np
from pp.component import Component
from pp.component import ComponentReference
from pp.ports.utils import select_electrical_ports

IDENTITY = (0, False)
R90 = (90, False)
R180 = (180, False)
R270 = (270, False)
mirror_x = (0, True)
mirror_y = (180, True)
mirror_x270 = (270, True)
mirror_x90 = (90, True)

TRANSFORMATION_MAP = {
    IDENTITY: "None",
    R90: "R90",
    R180: "R180",
    R270: "R270",
    mirror_y: "mirror_y",
    mirror_x: "mirror_x",
    mirror_x270: "mirror_x270",
    mirror_x90: "mirror_x90",
}

STR_TO_TRANSFORMATION_MAP = {v: k for k, v in TRANSFORMATION_MAP.items()}


def get_elec_ports_from_component_names(component, names=[]):
    """
    Args
        component <pp.Component>; should have component.info["components"]
    """
    e_ports = {}

    for name in names:
        _ports = {
            "{}_{}".format(name, p.name): p
            for p in select_electrical_ports(
                component.info["components"][name]
            ).values()
        }
        e_ports.update(_ports)

    # update port names
    for pname, p in e_ports.items():
        p.name = pname

    return e_ports


def gen_sref(component, transformation_name, port_name, position):
    """
    """

    if transformation_name not in TRANSFORMATION_MAP.values():
        raise ValueError(
            "{} is not a valid transformation({})".format(
                transformation_name, list(TRANSFORMATION_MAP.values())
            )
        )
    rotation_angle, x_reflection = STR_TO_TRANSFORMATION_MAP[transformation_name]
    position = np.array(position)

    if port_name is None:
        port_position = np.array([0, 0])
    else:
        if port_name not in component.ports.keys():
            raise ValueError(
                "{} port name not in {} ports ({})".format(
                    port_name, component.name, component.ports.keys()
                )
            )
        port_position = component.ports[port_name].midpoint

    device_ref = ComponentReference(device=component, origin=(0, 0))

    if x_reflection:  # Vertical mirror: Reflection across x-axis
        y0 = port_position[1]
        device_ref.reflect(p1=(0, y0), p2=(1, y0))

    if rotation_angle != 0:
        device_ref.rotate(rotation_angle, center=port_position)

    translation = np.array(position - port_position)

    device_ref.move(destination=translation)

    return device_ref


def netlist_to_component(components, connections, ports_map, position=(0, 0)):
    """
    Args:
        components:
            list of (component_id <str>, component <Component>, transform <tuple>)

        connections:
            list of (component_id1, port_name1, component_id2, port_name2)
            Has to be ordered such that any component refered to in the connections
            has already been placed (except for the first component)

        ports_map:
            {port_name: (component_id, port_name)}

    Returns: component
    the component has component.netlist

    [
        {
            "name": "CP2x2",
            "rank": 0,
            "ports": ["in1", "in2", "out1", "out2"],
            "type": "CPBIDIR",
            "settings": {"R": 0.5},
        },
        {
            "name": "WG_CAVITY",
            "rank": 0,
            "ports": ["in", "out"],
            "type": "WG",
            "settings": {"length": 50, "loss_dB_m": 60000, "width": 0.5},
        },
        {
            "name": "ring_resonator",
            "rank": 1,
            "type": "COMPOUND",
            "settings": {},
            "components": [
                "CP1, CP2x2, None, None, 0.0, 0.0",
                "WG1, WG_CAVITY, None, None, 0.0, 0.0",
            ],
            "connections": ["CP1, out1, WG1, in", "WG1, out, CP1, in1"],
            "ports": {"in1": "CP1, in2", "out1": "CP1, out2"},
        },
    ]

    mirror, rotation, x, y
    """
    if len(connections) == 0:
        raise ValueError(
            "Error number of connections", len(connections), len(components)
        )

    component_id, cmp_port, _, _ = connections[0]

    component, transform_name = components[component_id]

    # First component reference
    sref_start = gen_sref(component, transform_name, cmp_port, position)
    cmp_name_to_sref = {component_id: sref_start}

    # Iterate over all connections: create and place components
    for cmp1_name, port1, cmp2_name, port2 in connections:
        if cmp1_name not in cmp_name_to_sref:
            cmp1_name, port1, cmp2_name, port2 = cmp2_name, port2, cmp1_name, port1

        if cmp2_name not in cmp_name_to_sref:
            component, transform_name = components[cmp2_name]

            _ref = cmp_name_to_sref[cmp1_name]
            try:
                position = _ref.ports[port1].midpoint
            except Exception:
                print("{} has not port {}".format(cmp1_name, port1))

            sref = gen_sref(component, transform_name, port2, position)

            cmp_name_to_sref[cmp2_name] = sref

    c = Component()
    c.add(list(cmp_name_to_sref.values()))
    for port_name, (cmp_id, internal_port_name) in ports_map.items():
        c.add_port(port_name, port=cmp_name_to_sref[cmp_id].ports[internal_port_name])

    # Set aliases
    for cmp_id, ref in cmp_name_to_sref.items():
        c[cmp_id] = ref

    c.info["components"] = cmp_name_to_sref
    # c.components = components
    # c.connections = connections
    # c.ports_map = ports_map
    # c.cells = {
    #     cname: c for cname, (c, transf) in components.items()
    # }  # "CP1": (cpl, "None"),

    # add leaf cells to netlist
    netlist = []
    for name, (component, _) in components.items():
        netlist.append(
            dict(
                name=name,
                rank=0,
                ports=list(component.ports.keys()),
                settings=component.settings,
            )
        )
    # add the compound cell to the netlist
    netlist.append(
        dict(
            name="component_name",
            rank=1,
            type="COMPOUND",
            settings={},
            connections=[", ".join(i) for i in connections],
            ports={k: ", ".join(v) for k, v in ports_map.items()},
        )
    )
    c.netlist = netlist
    return c


if __name__ == "__main__":
    import pp
    from pp.components.ring_single_bus import ring_single_bus_netlist

    components, connections, ports_map = ring_single_bus_netlist()
    c = netlist_to_component(components, connections, ports_map)
    print(c.netlist)
    pp.show(c)



import numpy as np
from pp.geo_utils import path_length
from pp.routing.manhattan import remove_flat_angles
from pp.routing.manhattan import _is_horizontal
from pp.routing.manhattan import _is_vertical


def path_length_matched_points(
    list_of_waypoints, modify_segment_i=-2, bend_radius=10.0, dL0=0.0, nb_loops=1
):
    """
    Several types of paths won't match correctly.
    We do not try to handle all the corner cases here.
    If the paths are not well behaved enough, the input list_of_waypoints needs to be modified.

    Args:
        list_of_waypoints:  [[p1, p2, p3,...], [q1, q2, q3,...], ...] the number of turns have to be identical (usually means same number of points. exception is if there are some flat angles)
        bend_radius: used to estimate the position of new waypoints to accommodate bends with a given radius
        margin: some extra space to budget for in addition to the bend radius in most cases, the default is fine
        dL0: distance added to all path length compensation.  Useful is we want to add space for extra taper on all branches
        modify_segment_i: index of the segment which accomodates the new turns default is next to last segment

        nb_loops: number of extra loops added in the path
            if nb_loops==0, no extra loop is added, instead, in each route,
            the segment indexed by `modify_segment_i` is elongated to match
            the longuest route in `list_of_waypoints`

    returns: another list of waypoints where
        - the path_lenth of each waypoints list are identical
        - the number of turns are identical

    """

    common_params = {
        "list_of_waypoints": list_of_waypoints,
        "modify_segment_i": modify_segment_i,
        "dL0": dL0,
    }

    if nb_loops >= 1:
        return path_length_matched_points_add_waypoints(
            **common_params, bend_radius=bend_radius, nb_loops=nb_loops
        )
    else:
        return path_length_matched_points_modify_segment(**common_params)


def path_length_matched_points_modify_segment(list_of_waypoints, modify_segment_i, dL0):
    if type(list_of_waypoints) != list:
        raise ValueError(
            "list_of_waypoints should be a list, got {}".format(type(list_of_waypoints))
        )
    list_of_waypoints = [remove_flat_angles(l) for l in list_of_waypoints]
    lengths = [path_length(l) for l in list_of_waypoints]
    L0 = max(lengths)

    N = len(list_of_waypoints[0])

    # Find how many turns there are per path
    nb_turns = [len(l) - 2 for l in list_of_waypoints]

    # The paths have to have the same number of turns, otherwise this algo
    # cannot path length match
    if min(nb_turns) != max(nb_turns):
        raise ValueError(
            "Number of turns in paths have to be identical got \
        {}".format(
                nb_turns
            )
        )

    if modify_segment_i < 0:
        modify_segment_i = modify_segment_i + N + 1

    list_new_waypoints = []

    # For each list of waypoints, modify one segment in-place
    for i, waypoints in enumerate(list_of_waypoints):
        p_s0, p_s1, p_next = waypoints[modify_segment_i - 1 : modify_segment_i + 2]

        p_s0 = np.array(p_s0)
        p_s1 = np.array(p_s1)

        L = lengths[i]

        # Path length compensation length
        dL = (L0 - L) / 2

        # Additional fixed length
        dL = dL + dL0

        # Modify the segment to accomodate for path length matching
        # Two cases: vertical or horizontal segment
        if _is_vertical(p_s0, p_s1):
            sx = np.sign(p_next[0] - p_s1[0])

            dx = -sx * dL
            dp = (dx, 0)
            # Sequence of displacements to apply

        elif _is_horizontal(p_s0, p_s1):
            sy = np.sign(p_next[1] - p_s1[1])

            dy = -sy * dL
            dp = (0, dy)

        waypoints[modify_segment_i - 1] = p_s0 + dp
        waypoints[modify_segment_i] = p_s1 + dp

        list_new_waypoints += [waypoints]
    return list_new_waypoints


def path_length_matched_points_add_waypoints(
    list_of_waypoints,
    modify_segment_i=-2,
    bend_radius=10.0,
    margin=0.5,
    dL0=0.0,
    nb_loops=1,
):

    """
    Args:
        list_of_waypoints: a list of list_of_points:
            [[p1, p2, p3,...], [q1, q2, q3,...], ...]
            - the number of turns have to be identical
                (usually means same number of points. exception is if there are
                some flat angles)

        bend_radius: used to estimate the position of new waypoints to accommodate
                    bends with a given radius

        margin: some extra space to budget for in addition to the bend radius
            in most cases, the default is fine

        dL0: distance added to all path length compensation.
            Useful is we want to add space for extra taper on all branches

        modify_segment_i: index of the segment which accomodates the new turns
            default is next to last segment

        nb_loops: number of extra loops added in the path

    returns:
        another list of waypoints where:
            - the path_lenth of each waypoints list are identical
            - the number of turns are identical

    Several types of paths won't match correctly. We do not try to handle
    all the corner cases here. If the paths are not well behaved enough,
    the input list_of_waypoints needs to be modified.

    """

    print(list_of_waypoints[0])

    if type(list_of_waypoints) != list:
        raise ValueError(
            "list_of_waypoints should be a list, got {}".format(type(list_of_waypoints))
        )
    list_of_waypoints = [remove_flat_angles(l) for l in list_of_waypoints]
    lengths = [path_length(l) for l in list_of_waypoints]
    L0 = max(lengths)

    N = len(list_of_waypoints[0])

    # Find how many turns there are per path
    nb_turns = [len(l) - 2 for l in list_of_waypoints]

    # The paths have to have the same number of turns, otherwise cannot path-length
    # match with this algorithm
    if min(nb_turns) != max(nb_turns):
        raise ValueError(
            "Number of turns in paths have to be identical got \
        {}".format(
                nb_turns
            )
        )

    # To have flexibility in the path length, we need to add 4 bends
    """
    One path has to be converted in this way:

                      ----
                      |  |
                      |  |  This length is adjusted to make all path with the same length
                      |  |
    --------  ===> ---|  |---
    """

    # Get the points for the segment we need to modify
    a = margin + bend_radius
    if modify_segment_i < 0:
        modify_segment_i = modify_segment_i + N + 1
    list_new_waypoints = []

    for i, waypoints in enumerate(list_of_waypoints):
        p_s0, p_s1, p_next = waypoints[modify_segment_i - 2 : modify_segment_i + 1]

        p_s1 = np.array(p_s1)

        L = lengths[i]

        # Path length compensation length
        dL = (L0 - L) / (2 * nb_loops)

        # Additional fixed length
        dL = dL + dL0

        # Generate a new sequence of points which will replace this segment
        # Two cases: vertical or horizontal segment
        if _is_vertical(p_s0, p_s1):
            sx = np.sign(p_next[0] - p_s1[0])
            sy = np.sign(p_s1[1] - p_s0[1])

            dx = sx * (2 * a + dL)
            dy = sy * 2 * a

            # First new point to insert
            q0 = p_s1 + (0, -2 * nb_loops * dy)

            # Sequence of displacements to apply
            seq = [(dx, 0), (0, dy), (-dx, 0), (0, dy)] * nb_loops
            seq.pop()  # Remove last point to avoid flat angle with next point

        elif _is_horizontal(p_s0, p_s1):
            sy = np.sign(p_next[1] - p_s1[1])
            sx = np.sign(p_s1[0] - p_s0[0])

            dx = sx * 2 * a
            dy = sy * (2 * a + dL)

            q0 = p_s1 + (-2 * dx * nb_loops, 0)
            # Sequence of displacements to apply
            seq = [(0, dy), (dx, 0), (0, -dy), (dx, 0)] * nb_loops
            seq.pop()  # Remove last point to avoid flat angle with next point

        # Generate points to insert
        qs = [q0]
        for dp in seq:
            qs += [qs[-1] + dp]

        # print()
        # print(nb_loops)
        # for q in qs:
        # print(q)
        # print()
        inserted_points = np.stack(qs, axis=0)
        waypoints = np.array(waypoints)

        # Insert the points
        new_points = np.vstack(
            [
                waypoints[: modify_segment_i - 1],
                inserted_points,
                waypoints[modify_segment_i - 1 :],
            ]
        )
        list_new_waypoints += [new_points]

    return list_new_waypoints
